import os
import re
import sys
from ast import literal_eval as make_tuple
import configparser
import warnings
import importlib
import multiprocessing as mp
from typing import List, Optional, Union, Dict, Literal, Callable, Tuple, TypedDict, Any
from typing_extensions import Annotated
from types import ModuleType
import tempfile

import numpy as np
import xarray as xr
from numpy.typing import ArrayLike

from pydantic import (
    BaseModel, Field, computed_field, field_validator, model_validator, 
    ConfigDict, TypeAdapter, ValidationError, validator
)
from pydantic.functional_validators import BeforeValidator, AfterValidator
from pydantic.functional_serializers import PlainSerializer

import pymob
from pymob.utils.store_file import scenario_file, converters
from pymob.sim.parameters import Param, NumericArray, OptionRV
# this loads at the import of the module
default_path = sys.path.copy()

# class FloatParam(BaseModel):
#     name: Optional[str] = None
#     value: float = 0.0
#     min: Optional[float] = None
#     max: Optional[float] = None
#     step: Optional[float] = None
#     prior: Optional[str] = None
#     free: bool = True


# class ArrayParam(BaseModel):
#     name: Optional[str] = None
#     value: List[float] = [0.0]
#     min: Optional[List[float]] = None
#     max: Optional[List[float]] = None
#     step: Optional[List[float]] = None
#     prior: Optional[str] = None
#     free: bool = True

class PymobModel(BaseModel):
    def __getitem__(self, key: str):
        """Allow getting attribute by key (like a dict)"""
        return getattr(self, key)

    def __setitem__(self, key: str, value):
        """Allow setting attribute by key (like a dict), with validation"""
        if key not in self.__annotations__:
            raise KeyError(f"Key '{key}' not found")

        # Set the value to the attribute
        setattr(self, key, value)

        # Trigger Pydantic validation, like when setting the attribute directly
        try:
            # Re-validate the model instance after setting the value
            self.__class__.parse_obj(self.dict())
        except ValidationError as e:
            raise ValueError(f"Validation failed: {e}")

class ModelParameterDict(TypedDict):
    parameters: Dict[str, float|str|int]
    y0: xr.Dataset
    x_in: xr.Dataset

class DataVariable(BaseModel):
    """Describe a data variable

    Parameters
    ----------
    dimensions: List[str]
        Specifies the dimensions, which have to be found in the observations and 
        the dimensional order of the data variable
    
    min: float
        The possible minimum of a data variable, is used to construct scalers.
        Defaults to 'nan', which sets the minimum to the minimum of the 
        observations

    min: float
        The possible maxmimum of a data variable, is used to construct scalers
        Defaults to 'nan', which sets the maximum to the maximum of the 
        observations

    observed: bool
        If the data-variable was observed or not. Defaults to True

    dimensions_evaluator: List[str]
        Specifies the dimensions and their order returned by the evaluator.
        This is necessary to bring observations and results together, if for some
        reason the returned simulation results from the evaluator have a different
        dimensional order than the observations

    Returns
    -------
    None

    """
    model_config = ConfigDict(
        validate_assignment=True,
        extra="forbid"
    )
    dimensions: List[str]
    min: float = np.nan
    max: float = np.nan
    observed: bool = True
    dimensions_evaluator: Optional[List[str]] = None

    @model_validator(mode="after")
    def post_update(self):
        if self.dimensions_evaluator is not None:
            if len(self.dimensions_evaluator) != len(self.dimensions):
                self.dimensions_evaluator = None

        return self
        
        
    @field_validator("dimensions_evaluator", mode="after")
    def set_data_variable_bounds(cls, v, info, **kwargs):
        # For conditionally updating values (e.g. when data variables change)
        # see https://github.com/pydantic/pydantic/discussions/7127
        dimensions = info.data.get("dimensions")
        if v is not None:
            if len(v) != len(dimensions):
                raise AssertionError(
                    f"Evaluator dimensions {v} must have the "
                    f"same length as observations {dimensions} dimensions."
                )

            elif set(v) != set(dimensions):
                raise AssertionError(
                    f"Evaluator dimensions {set(v)} must have the "
                    f"same names as observations {set(dimensions)} dimensions."
                )

            else:
                return v
        else:
            return None
    
    def __eq__(self, value):
        smin, smax = self.min, self.max
        vmin, vmax = value.min, value.max

        if np.all(np.isnan([smin, vmin])):
            min_equal = True
        else:
            min_equal = all(vmin == smax)
        
        if np.all(np.isnan([smax, vmax])):
            max_equal = True
        else:
            max_equal = all(vmax == smax)
            
        return all([
            self.dimensions == value.dimensions,
            self.observed == value.observed,
            self.dimensions_evaluator == value.dimensions_evaluator,
            min_equal,
            max_equal,
        ])

class ParameterDict(dict):
    def __init__(self, *args, **kwargs):
        self.callback: Callable = kwargs.pop('callback', None)
        super().__init__(*args, **kwargs)
    
    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        if self.callback:
            self.callback(self)
    
    def update(self, *args, **kwargs):
        super().update(*args, **kwargs)
        if self.callback:
            self.callback(self)

def string_to_list(option: Union[List, str]) -> List:
    if isinstance(option, (list, tuple)):
        return list(option)
    
    if len(option) == 0:
        return []
    elif " " not in option:
        return [option] 
    else:
        return [i.strip() for i in option.split(" ")]


def string_to_tuple(option: Union[List, str]) -> Tuple:
    return tuple(string_to_list(option))


def string_to_dict(
        option: Union[Dict[str,str|float|int|List[float|int|str]], str]
    ) -> Dict[str,str|float|int|List[float|int|str]]:
    """Expects a string of this form 
    e.g. 'value=1 max=10 min=0 prior=Lognormal(loc=2,scale=1)'
    """
    if isinstance(option, Dict):
        return option
    
    retdict = {}
    if len(option) == 0:
        return retdict

    for i in option.split(" "):
        k, v = i.strip().split(sep="=", maxsplit=1)
        parsed = False
        if not parsed:
            try:
                parsed_value = TypeAdapter(float).validate_json(v)
                parsed = True
            except ValidationError:
                pass

        if not parsed:
            try:
                # v_ = np.array(ast.literal_eval(v))
                # _cfg = ConfigDict(arbitrary_types_allowed=True)
                parsed_value = TypeAdapter(NumericArray).validate_json(v)
                parsed = True
            except ValueError:
                pass

        if not parsed and v[0] == "(" and v[-1] == ")":
            try:
                parsed_value = make_tuple(v)
                parsed = True
            except ValueError:
                pass

        # TODO: This expression seems to be wrong, but it causes no errors
        if not parsed and  v[0]=="[" and v[-1]=="]":
            try:
                v_ = v.strip("[]").split(",")
                # remove double quotes
                v_ = [re.sub(r'^"|"$', '', s) for s in v_]
                v_ = [re.sub(r"^'|'$", '', s) for s in v_]
                v_ = [v_i for v_i in v_ if v_i != ""]
                parsed_value = TypeAdapter(List[str]).validate_python(v_)
                parsed = True
            except ValidationError:
                pass

        if not parsed:
            parsed_value = v

        retdict.update({k:parsed_value})
            
    return retdict


def string_to_param(option:str|Param) -> Param:
    if isinstance(option, Param):
        return option
    else:
        param_dict = string_to_dict(option)
        return Param.model_validate(param_dict, strict=False)
        

def string_to_datavar(option:str|DataVariable) -> DataVariable:
    if isinstance(option, DataVariable):
        return option
    else:
        param_dict = string_to_dict(option)
        return DataVariable.model_validate(param_dict, strict=False)
        

def dict_to_string(dct: Dict, replace_whitespace=""):
    string_items = []
    for k, v in dct.items():
        if isinstance(v, np.ndarray):
            v = v.tolist()

        expr = f"{k}={v}".replace(" ", replace_whitespace)
        string_items.append(expr)

    return " ".join(string_items)


def list_to_string(lst: List):
    return " ".join([str(l).replace(" ", "") for l in lst])


def param_to_string(prm: Param):
    return dict_to_string(prm.model_dump(exclude_none=True, mode="json"))

def datavar_to_string(prm: DataVariable):
    return dict_to_string(prm.model_dump(exclude_none=True))


serialize_list_to_string = PlainSerializer(
    list_to_string, 
    return_type=str, 
    when_used="json"
)


serialize_dict_to_string = PlainSerializer(
    dict_to_string, 
    return_type=str, 
    when_used="json"
)


serialize_param_to_string = PlainSerializer(
    param_to_string, 
    return_type=str, 
    when_used="json"
)

serialize_datavar_to_string = PlainSerializer(
    datavar_to_string, 
    return_type=str, 
    when_used="json"
)

to_str = PlainSerializer(lambda x: str(x), return_type=str, when_used="json")


OptionListStr = Annotated[
    List[str], 
    BeforeValidator(string_to_list), 
    serialize_list_to_string
]

OptionTupleStr = Annotated[
    Tuple[str, ...], 
    BeforeValidator(string_to_tuple), 
    serialize_list_to_string
]

OptionDictStr = Annotated[
    Dict[str,str|float|int|List[float|int]], 
    BeforeValidator(string_to_dict), 
    serialize_dict_to_string
]

OptionDataVariable = Annotated[
    DataVariable, 
    BeforeValidator(string_to_datavar), 
    serialize_datavar_to_string
]

OptionParam = Annotated[
    Param, 
    BeforeValidator(string_to_param), 
    serialize_param_to_string
]


# OptionListFloat = Annotated[
#     List[float], 
#     BeforeValidator(string_to_list), 
#     serialize_list_to_string
# ]

        
class Casestudy(PymobModel):
    model_config = {"validate_assignment" : True}
    init_root: str = Field(default=os.getcwd(), exclude=True)
    root: str = "."

    name: str = "unnamed_case_study"
    version: Optional[str] = None
    pymob_version: Optional[str] = None
    scenario: str = "unnamed_scenario"
    package: str = "case_studies"
    modules: OptionListStr = ["sim", "mod", "prob", "data", "plot"]
    simulation: str = Field(default="Simulation", description="Simulation Class defined in sim.py module in the case study.")

    output: Optional[str] = None
    data: Optional[str] = None
    scenario_path_override: Optional[str] = None

    observations: Optional[str] = None
    
    logging: str = "DEBUG"
    logfile: Optional[str] = None

    @computed_field
    @property
    def output_path(self) -> str:
        if self.output is not None:
            return self.output
        else:
            return os.path.join(
                os.path.relpath(self.package),
                os.path.relpath(self.name),
                "results",
                self.scenario,
            )
    
    @output_path.setter
    def output_path(self, value) -> None:
        self.output = value

    @computed_field
    @property
    def data_path(self) -> str:
        if self.data is not None:
            return self.data
        else:
            return os.path.join(
                os.path.relpath(self.package), 
                os.path.relpath(self.name),
                "data"
            )
        
    @data_path.setter
    def data_path(self, value) -> None:
        self.data = value
    
    @computed_field
    @property
    def default_settings_path(self) -> str:
        return os.path.join(
            os.path.relpath(self.package),
            os.path.relpath(self.name),
            "scenarios", 
            self.scenario, 
            "settings.cfg"
        )
        
    @property
    def scenario_path(self):
        if self.scenario_path_override is not None:
            return self.scenario_path_override

        package_path = os.path.basename(os.path.abspath(self.package))
        if package_path == self.name:
            # if the package path is identical to the case study path, then 
            # the scenario is directly located in the package.
            return os.path.join(
                os.path.relpath(self.package), 
                "scenarios", 
                self.scenario
            )
        else:
            return os.path.join(
                os.path.relpath(self.package), 
                os.path.relpath(self.name),
                "scenarios", 
                self.scenario
            )
    
    @field_validator("root", mode="after")
    def set_root(cls, new_value, info, **kwargs):
        # For conditionally updating values (e.g. when data variables change)
        # see https://github.com/pydantic/pydantic/discussions/7127
        os.chdir(info.data.get("init_root"))  # this resets the root to the original value
        package = info.data.get("package")
        name = info.data.get("name")
        root = os.path.abspath(new_value)
        if root != os.getcwd():
            if not os.path.exists(os.path.join(root, package)):
                raise FileNotFoundError(
                    f"Case study collection-directory '{package}' does not "
                    f"exist in {root}. If the root is the case study. Set "
                    f"sim.config.package = '.'"
                )
            else:
                if not os.path.exists(os.path.join(root, package, name)):
                    raise FileNotFoundError(
                        f"Case study '{name}' "
                        f"does not exist in {os.path.join(root, package)}. "
                    )   
                else:
                    os.chdir(root)
                    print(f"Working directory: '{root}'.")
                    return root
        else:
            print(f"Working directory: '{root}'.")
            return root




class Simulation(PymobModel):
    model_config = {"validate_assignment" : True, "extra": "allow"}

    model: Optional[str] = Field(default=None, validate_default=True, description="The deterministic model")
    solver: Optional[str] = Field(default=None, validate_default=True)
    
    y0: OptionListStr = []
    x_in: OptionListStr = []

    input_files: OptionListStr = []
    # data_variables: OptionListStr = []
    n_ode_states: int = -1
    batch_dimension: str = "batch_id"
    x_dimension: str = "time"
    modeltype: Literal["stochastic", "deterministic"] = "deterministic"
    solver_post_processing: Optional[str] = Field(default=None, validate_default=True)
    seed: Annotated[int, to_str] = 1

class Datastructure(PymobModel):
    __pydantic_extra__: Dict[str,OptionDataVariable]
    model_config = ConfigDict(extra="allow", validate_assignment=True)

    def remove(self, key) -> None:
        """Removes a data variable from the data structure"""
        if key not in self.__pydantic_extra__:
            warnings.warn(
                f"'{key}' is not a data-variable. Data variables are: "
                f"{self.data_variables}."
            )
            return
        
        deleted_var = self.__pydantic_extra__.pop(key)
        print(f"Deleted '{key}' DataVariable({deleted_var}).")

    def __getitem__(self, key):
        return self.__pydantic_extra__[key]

    @property
    def data_variables(self) -> List[str]:
        return [k for k in self.__pydantic_extra__.keys()]
    
    @property
    def observed_data_variables(self) -> List[str]:
        return [k for k, v in self.__pydantic_extra__.items() if v.observed]
    
    @property
    def dimensions(self) -> List[str]:
        # TODO: Remove when dimensions is not accessed any longer
        warnings.warn(
            "Legacy method, will be deprecated soon. This works only if all "
            "Data variables have the same dimension",
            category=DeprecationWarning
        )
        dims = []
        for k, v in self.__pydantic_extra__.items():
            for d in v.dimensions:
                if d not in dims:
                    dims.append(d)
        return dims
    
    @property
    def all(self) -> Dict[str, DataVariable]:
        return {k: v for k, v in self.__pydantic_extra__.items()}

    @property
    def dimdict(self) -> Dict[str, List[str]]:
        return {k: v.dimensions for k, v in self.__pydantic_extra__.items()}

    @property
    def observed_dimdict(self) -> Dict[str, List[str]]:
        return {k: v.dimensions for k, v in self.__pydantic_extra__.items() if v.observed}


    @property
    def var_dim_mapper(self) -> Dict[str, List[str]]:
        var_dim_mapper = {}
        for k, v in self.all.items():
            if v.dimensions_evaluator is None:
                dims_evaluator = v.dimensions
            else:
                dims_evaluator = v.dimensions_evaluator

            var_dim_mapper.update({
                k: [v.dimensions.index(e_i) for e_i in dims_evaluator]
            })

        return var_dim_mapper
    
    @property
    def evaluator_dim_order(self) -> List[str]:
        # TODO: Remove when dimensions is not accessed any longer
        warnings.warn(
            "Legacy method, will be deprecated soon. This works only if all "
            "Data variables have the same dimension",
            category=DeprecationWarning
        )
        dims = []
        for k, v in self.all.items():
            if v.dimensions_evaluator is None:
                dims_evaluator = v.dimensions
            else:
                dims_evaluator = v.dimensions_evaluator

            for d in dims_evaluator:
                if d not in dims:
                    dims.append(d)
        return dims
    
    @property
    def data_variables_min(self):
        return [v.min for v in self.__pydantic_extra__.values()]

    @property
    def data_variables_max(self):
        return [v.max for v in self.__pydantic_extra__.values()]

    @property
    def observed_data_variables_min(self):
        return [v.min for v in self.__pydantic_extra__.values() if v.observed]

    @property
    def observed_data_variables_max(self):
        return [v.max for v in self.__pydantic_extra__.values() if v.observed]

class Solverbase(PymobModel):
    model_config = ConfigDict(
        validate_assignment=True, 
        extra="forbid"
    )
    x_dim: str = "time"
    exclude_kwargs_model: OptionTupleStr = ("t", "time", "x_in", "y", "x", "Y", "X")
    exclude_kwargs_postprocessing: OptionTupleStr = ("t", "time", "interpolation", "results")

class Jaxsolver(PymobModel):
    diffrax_solver: str = "Dopri5"
    rtol: float = 1e-6
    atol: float = 1e-7
    pcoeff: float = 0.0
    icoeff: float = 1.0
    dcoeff: float = 0.0
    max_steps: int = int(1e5)
    throw_exception: bool = True

class Inference(PymobModel):
    model_config = {"validate_assignment" : True}

    eps: float = 1e-8
    objective_function: str = "total_average"
    n_objectives: Annotated[int, to_str] = 1
    objective_names: OptionListStr = []
    backend: Optional[str] = None
    extra_vars: OptionListStr = []
    plot: Optional[str|Callable] = None
    n_predictions: Annotated[int, to_str] = 100

class Multiprocessing(PymobModel):
    _name = "multiprocessing"
    model_config = ConfigDict(validate_assignment=True, extra="ignore")

    # TODO: Use as private field
    cores: Annotated[int, to_str] = 1
    
    @property
    def n_cores(self) -> Annotated[int, to_str]:
        cpu_avail = mp.cpu_count()
        cpu_set = self.cores
        if cpu_set <= 0:
            return cpu_avail + cpu_set
        else: 
            return cpu_set
        
class Modelparameters(PymobModel):
    __pydantic_extra__: Dict[str,OptionParam]
    model_config = ConfigDict(extra="allow", validate_assignment=True)

    @property
    def all(self) -> Dict[str,OptionParam]:
        return self.__pydantic_extra__
    
    @all.setter
    def all(self, value):
        self.__pydantic_extra__ = value

    @property
    def free(self) -> Dict[str,OptionParam]:
        return {k:v for k, v in self.all.items() if v.free}

    @property
    def fixed(self) -> Dict[str,OptionParam]:
        return {k:v for k, v in self.all.items() if not v.free}

    @property
    def n_free(self) -> int:
        return len(self.free)
    
    @property
    def free_value_dict(self) -> Dict[str,float|NumericArray]:
        return {k:v.value for k, v in self.free.items()}
    
    @property
    def fixed_value_dict(self) -> Dict[str,float|NumericArray]:
        return {k:v.value for k, v in self.fixed.items()}
    
    @property
    def value_dict(self) -> Dict[str,float|NumericArray]:
        return {k:v.value for k, v in self.all.items()}
    
    @property
    def dimensions(self) -> List[str]:
        # TODO: Remove when dimensions is not accessed any longer
        warnings.warn(
            "Legacy method, will be deprecated soon. This works only if all "
            "Data variables have the same dimension",
            category=DeprecationWarning
        )
        dims = []
        for k, v in self.all.items():
            for d in v.dims:
                if d not in dims:
                    dims.append(d)
        return dims

    def remove(self, key) -> None:
        """Removes a Parameter"""
        if key not in self.all:
            warnings.warn(
                f"'{key}' is not a parameter. Parameters are: "
                f"{list(self.all.keys())}."
            )
            return
        
        deleted_par = self.all.pop(key)
        print(f"Deleted '{key}' Param({deleted_par}).")

    def reorder(self, keys: List[str]):
        """Reorders model parameters. This may be necessary for hierarchical 
        models, because priors take draws from hyperpriors to parameterize their
        distributions. Hence, they must be available earlier.

        Parameters
        ----------

        keys : List[str]
            A list of model parameters to sort the model_parameter dictionary 
            after. If the keys list is smaller than the list of model parameters,
            unlisted parameters will be appended to the keys list in order.
        """
        if len(keys) < len(self.all):
            keys = keys + [k for k in self.all.keys() if k not in keys]
            
        self.all = {k:self.all[k] for k in keys} 

class Errormodel(PymobModel):
    __pydantic_extra__: Dict[str,OptionRV]
    model_config = ConfigDict(extra="allow", validate_assignment=True)
    
    @property
    def all(self) -> Dict[str,OptionRV]:
        return self.__pydantic_extra__

class Pyabc(PymobModel):
    model_config = {"validate_assignment" : True}

    sampler: str = "SingleCoreSampler"
    population_size: Annotated[int, to_str] = 100
    minimum_epsilon: Annotated[float, to_str] = 0.0
    min_eps_diff: Annotated[float, to_str] = 0.0
    max_nr_populations: Annotated[int, to_str] = 1000
    
    # database configuration
    database_path: str = f"{tempfile.gettempdir()}/pyabc.db"

class Redis(PymobModel):
    model_config = {"validate_assignment" : True, "protected_namespaces": ()}

    _name = "inference.pyabc.redis"

    # redis configuration
    password: str = "nopassword"
    port: Annotated[int, to_str] = 1111

    # eval configuration
    n_predictions: Annotated[int, to_str] = Field(default=50, alias="eval.n_predictions")
    history_id: Annotated[int, to_str] = Field(default=-1, alias="eval.history_id")
    model_id: Annotated[int, to_str] = Field(default=0, alias="eval.model_id")


class Pymoo(PymobModel):
    model_config = ConfigDict(validate_assignment=True, extra="ignore")

    algortihm: str = "UNSGA3"
    population_size: Annotated[int, to_str] = 100
    max_nr_populations: Annotated[int, to_str] = 1000
    ftol: Annotated[float, to_str] = 1e-5
    xtol: Annotated[float, to_str] = 1e-7
    cvtol: Annotated[float, to_str] = 1e-7
    verbose: Annotated[bool, to_str] = True
    
class Numpyro(PymobModel):
    model_config = ConfigDict(validate_assignment=True, extra="ignore")
    user_defined_probability_model: Optional[str] = None
    user_defined_error_model: Optional[str] = None
    user_defined_preprocessing: Optional[str] = None
    gaussian_base_distribution: bool = False
    
    # inference arguments
    kernel: str = "nuts"
    init_strategy: str = "init_to_uniform"
     
    # mcmc parameters
    chains: Annotated[int, to_str] = 1
    draws: Annotated[int, to_str] = 2000
    warmup: Annotated[int, to_str] = 1000
    thinning: Annotated[int, to_str] = 1
    
    # nuts arguments
    nuts_draws: Annotated[int, to_str] = 2000
    nuts_step_size: Annotated[float, to_str] = 0.8
    nuts_max_tree_depth: Annotated[int, to_str] = 10
    nuts_target_accept_prob: Annotated[float, to_str] = 0.8
    nuts_dense_mass: Annotated[bool, to_str] = True
    nuts_adapt_step_size: Annotated[bool, to_str] = True
    nuts_adapt_mass_matrix: Annotated[bool, to_str] = True

    # sa parameters
    sa_adapt_state_size: Optional[int] = None

    # svi parameters
    svi_iterations: Annotated[int, to_str] = 10_000
    svi_learning_rate: Annotated[float, to_str] = 0.0001

class Report(PymobModel):
    model_config = ConfigDict(validate_assignment=True, extra="ignore")
    
    debug_report: Annotated[bool, to_str] = False
    pandoc_output_format: Literal["html", "latex-si", "latex", "pdf"] = "html"

    model: Annotated[bool, to_str] = True
    parameters: Annotated[bool, to_str] = True
    parameters_format: Literal["xarray", "pandas"] = "pandas"
    
    diagnostics: Annotated[bool, to_str] = True
    diagnostics_with_batch_dim_vars: Annotated[bool, to_str] = False
    diagnostics_exclude_vars: OptionListStr = []
    
    goodness_of_fit: Annotated[bool, to_str] = True
    goodness_of_fit_use_predictions: Annotated[bool, to_str] = True
    goodness_of_fit_nrmse_mode: Literal["mean", "range", "iqr"] = "range"

    table_parameter_estimates: Annotated[bool, to_str] = True
    table_parameter_estimates_format: Literal["latex", "csv", "tsv"] = "csv"
    table_parameter_estimates_significant_figures: int = 3
    table_parameter_estimates_error_metric: Literal["hdi", "sd"] = "sd"
    table_parameter_estimates_parameters_as_rows: Annotated[bool, to_str] = True
    table_parameter_estimates_with_batch_dim_vars: Annotated[bool, to_str] = False
    table_parameter_estimates_exclude_vars: OptionListStr = []
    table_parameter_estimates_override_names: OptionDictStr = {}

    plot_trace: Annotated[bool, to_str] = True
    plot_parameter_pairs: Annotated[bool, to_str] = True

class Config(BaseModel):
    """Configuration manager for pymob."""
    model_config = {"validate_assignment" : True, "extra": "allow", "protected_namespaces": ()}
    _config: configparser.ConfigParser
    _modules: Dict[str,ModuleType]

    def __init__(
        self,
        config: Optional[Union[str, configparser.ConfigParser]] = None,
    ) -> None:

        _cfg_fp = None
        interp = configparser.ExtendedInterpolation()
        if isinstance(config, str):
            _config = configparser.ConfigParser(
                converters=converters,
                interpolation=interp
            )        
            _config.optionxform = str # type: ignore
            _cfg_file_paths = _config.read(config)
            try:
                _cfg_fp = _cfg_file_paths[0]
            except IndexError:
                raise FileNotFoundError(
                    f"Config file: {config} could not be found."
                ) 
        elif isinstance(config, configparser.ConfigParser):
            _config = config
        else:
            _config = configparser.ConfigParser(
                converters=converters,
                interpolation=interp
            )
            _config.optionxform = str # type: ignore

        # pass arguments to config

        if _cfg_fp is not None: _config.set("case-study", "settings_path", _cfg_fp)
        cfg_dict = {k:dict(s) for k, s in dict(_config).items() if k != "DEFAULT"}
        super().__init__(**cfg_dict)

        self._config = _config
        self._modules = {}

    case_study: Casestudy = Field(default=Casestudy(), alias="case-study")
    simulation: Simulation = Field(default=Simulation())
    data_structure: Datastructure = Field(default=Datastructure(), alias="data-structure") # type:ignore
    solverbase: Solverbase = Field(default=Solverbase())
    jaxsolver: Jaxsolver = Field(default=Jaxsolver(), alias="jax-solver")
    inference: Inference = Field(default=Inference())
    model_parameters: Modelparameters = Field(default=Modelparameters(), alias="model-parameters") #type: ignore
    error_model: Errormodel = Field(default=Errormodel(), alias="error-model") # type: ignore
    multiprocessing: Multiprocessing = Field(default=Multiprocessing())
    inference_pyabc: Pyabc = Field(default=Pyabc(), alias="inference.pyabc")
    inference_pyabc_redis: Redis = Field(default=Redis(), alias="inference.pyabc.redis")
    inference_pymoo: Pymoo = Field(default=Pymoo(), alias="inference.pymoo")
    inference_numpyro: Numpyro = Field(default=Numpyro(), alias="inference.numpyro")
    report: Report = Field(default=Report(), alias="report")
        
    @property
    def input_file_paths(self) -> list:
        paths_input_files = []
        for file in self.simulation.input_files:
            fp = scenario_file(file, self.case_study.name, self.case_study.scenario, pkg_dir=self.case_study.package)
            paths_input_files.append(fp)


        file = self.case_study.observations
        if file is None:
            return paths_input_files
        else:
            if not os.path.isabs(file):
                fp = os.path.join(self.case_study.data_path, file)
            else:
                fp = file
            paths_input_files.append(fp)

            return paths_input_files

    def print(self):
        print("Simulation configuration", end="\n")
        print("========================")
        for section, field_info in self.model_fields.items():
            print(f"{section}({getattr(self, section)})", end="\n") # type: ignore

        print("========================", end="\n")

    def save(self, fp: Optional[str]=None, force=False):
        """Saves the configuration to a settings.cfg file
        
        Uses serializers defined at the top, which parse the options to str
        so they can be processed by configfile. 

        In case the model configuration should be stored to a json file use
        something like `json.dumps(self.model_dump())`, because the build in
        function, is somewhat disabled by the listparsers which are needed for
        configfile lists.

        Parameters
        ----------
        fp: Optiona[str] file path to write the settings file to
        force: [bool] should the settings file be overwritten without asking
            for user confirmation (default: False)
        """

        settings = self.model_dump(
            by_alias=True, 
            mode="json", 
            exclude_none=True,
            exclude={"case_study": {"output_path", "data_path", "root", "init_root", "default_settings_path"}}
        )
        self._config.update(**settings)

        if fp is None:
            file_path = self.case_study.default_settings_path
        else:
            file_path = os.path.abspath(fp)

        write = True
        if os.path.exists(file_path) and not force:
            ui_overwrite = input("Settings file already exists. Overwrite? [y/N]")
            write = True if ui_overwrite == "y" else False
        else:
            # create a directory for the new scenario file
            os.makedirs(os.path.dirname(file_path), exist_ok=True)

        if write:
            with open(file_path, "w") as f:
                self._config.write(f)

    def create_directory(self, directory: Literal["results", "scenario"], force=False):
        if directory == "results":
            p = os.path.abspath(self.case_study.output_path)
        elif directory == "scenario":
            p = os.path.abspath(self.case_study.scenario_path)
        else:
            raise NotImplementedError(
                f"{directory.capitalize()} is not an expected directory in the "
                "case study logic. Use one of 'results' or 'scenario'."
            )

        if os.path.exists(p):
            print(f"{directory.capitalize()} directory exists at '{p}'.")
            return
        else:
            if not force:
                answer = input(f"Create {directory} directory at '{p}'? [Y/n]")
            else:
                answer = "y"

            if answer.lower() == "y" or answer.lower() == "":
                os.makedirs(p)
                print(f"{directory.capitalize()} directory created at '{p}'.")
            else:
                print(f"No {directory} directory created.")

    def import_casestudy_modules(self, reset_path=False):
        """
        this script handles the import of a case study without the typical 
        __init__.py file. It iterates through all .py files in the root directory
        of the case study (typically: sim, mod, stats, plot, data, prior)
        and imports them with import_module(...)
        """

        # potential BUG: This is not safe. It is not guaranteed that the 
        # case study has the same name as the package. But it might be in the future
        package = self.case_study.name

        if "-" in package or " " in package:
            warnings.warn(
                f"Case-study contained {package} contained unallowed "+
                "characters: ['-', ' ']. "+
                "The characters will be replaced with underscores ('_') for "+
                "importing the package modules. " +
                "In the future, the name of the case study should be the same as " +
                "as the package where the modules are located. This name must not "
                "contain hyphens ('-') or whitespace characters (' '),",
                category=UserWarning
            )
            _package = package.replace("-", "_").replace(" ", "_")
        else:
            _package = package

        spec = importlib.util.find_spec(_package)
        if spec is not None:
            for module in self.case_study.modules:
                try:
                    # TODO: Consider importing modules as a nested dictionary 
                    # with the indexing key being the package. The package
                    # cannot be derived from the class, if a method, that is 
                    # executed on a lower level case-study, should target that 
                    # a module belonging to the same package, because if the
                    # object is used, it would resolve to the package of the
                    # higher level case-study
                    m = importlib.import_module(f"{_package}.{module}")
                    self._modules.update({module: m})
                except ModuleNotFoundError:
                    warnings.warn(
                        f"Module {module}.py not found in {_package}."
                        f"Missing modules can lead to unexpected behavior. "
                        f"Does your case study have a {module}.py file? "
                        f"It should have the line `from PARENT_CASE_STUDY."
                        f"{module} import *` to import all objects from "
                        "the parent case study."
                    )
            return

        # reset the path to avoid importing modules form case-studies used
        # before in the same session
        if reset_path:
            # default path needs to be copied, otherwise it will be updated
            # when setting sys.path
            sys.path = default_path.copy()

        # append relevant paths to sys
        package = os.path.join(
            self.case_study.root, 
            self.case_study.package
        )
        if package not in sys.path:
            sys.path.insert(0, package)
            print(f"Inserted '{package}' in PATH at index=0")
    
        case_study = os.path.join(
            self.case_study.root, 
            self.case_study.package,
            self.case_study.name,
            # Account for package architecture 
            self.case_study.name
        )
        if case_study not in sys.path:
            sys.path.insert(0, case_study)
            print(f"Inserted '{case_study}' in PATH at index=0")

        for module in self.case_study.modules:
            # remove modules of a different case study that might have been
            # loaded in the same session.
            if module in sys.modules:
                _ = sys.modules.pop(module)

        for module in self.case_study.modules:
            try:
                m = importlib.import_module(module, package=case_study)
                self._modules.update({module: m})
            except ModuleNotFoundError:
                warnings.warn(
                    f"Module {module}.py not found in {case_study}."
                    f"Missing modules can lead to unexpected behavior."
                    "If a module is not imported, you can specify it in the "
                    "Config 'config.case_study.modules = [...]'"
                )

    def import_simulation_from_case_study(self):
        try:
            Simulation = getattr(self._modules["sim"], self.case_study.simulation)
        except:
            raise ImportError(
                f"Simulation class '{self.case_study.simulation}' "
                "could not be found. Make sure the simulaton option is spelled "
                "correctly or specify an class that exists in sim.py"
                "If you are using pymob to work on different case-studies in "
                "the same session, make sure to reset the path by "
                "using `import_casestudy_modules(reset_path=True)`"
            )
        
        return Simulation

    def set_option(self, section: str, option: str, value: str):
        sect = getattr(self, section)
        if isinstance(sect, Modelparameters):
            if (value == "" or value == "None") and option in sect.all:
                sect.remove(option)
            elif (value == "" or value == "None") and option not in sect.all:
                pass
            else:
                sect[option] = value
        else:
            sect[option] = value


import click

@click.command
@click.option("--file", "-f", type=str, nargs=1, help="Path to the config file (usually in scenario/.../settings.cfg)")
@click.option("--options", "-o", type=str, multiple=True, help="The option or options to configure. Combine sections and option like this 'simulation.seed=1' options that have spaces need to be wraped in quotes")
def configure(file, options: Tuple[str, ...]):
    config = Config(file)
    for opt in options:
        key, val = opt.split("=", 1)

        section, option = key.strip(" ").rsplit(".", 1)

        section = section.replace("-","_").replace(".","_")

        if section == "jax_solver":
            section == "jaxsolver"

        value = val.strip(" ")
        config.set_option(section, option, value)

    config.save(file, force=True)


import os
import click
import numpy as np
from matplotlib import pyplot as plt

# import own packages
from pymob.utils.store_file import (
    import_package, opt, prepare_casestudy, prepare_scenario, 
    store_sbi_simulations)
from pymob.utils import help

# Prepare priors and simulator =================================================
# training the network. Here I learn the relationship between parameter inout
# and simulation outputs

@click.command()
@click.option("-c", "--case_study", type=str, default="core_daphnia", 
              help=help.case_study)
@click.option("-s", "--scenario", type=str, default="expo_control", 
              help=help.scenario)
def main(case_study, scenario):
    config = prepare_casestudy((case_study, scenario), "sbi.cfg")

    # parse config file
    dataset_kwargs = {k:config["dataset"].getlist(k) for k in config["dataset"].keys()}
    output = config["case-study"].get("output")
    plot_output = config["case-study"].get("plots")

    mod = import_package(package_path=config["case-study"]["package"])
    summary_statistics = getattr(mod.stats, config["sbi"]["summary_stats"])    
    dataset = getattr(mod.data, config["sbi"]["dataset"])
    prior = getattr(mod.prior, config["sbi"]["prior"])
    # Load Estimator and samples (optional) ========================================
    # Plug in Real data ============================================================
    x_experiment = dataset(**dataset_kwargs)
    x_experiment = summary_statistics(x_experiment)
    # x_experiment = x_experiment[~x_experiment.isnan().any(1), :] # exclude rows with nan values

    stats_names = x_experiment.coords["variable"].values

    x_sim = np.loadtxt(os.path.join(output, "x.txt"))
    theta_sim = np.loadtxt(os.path.join(output, "theta.txt"))

    def plot_good_param_space(n_theta, n_summary_stat, xscale="linear"):
        t_s = theta_sim[:, n_theta]
        x_s = x_sim[:, n_summary_stat]
        x_e = x_experiment[:, n_summary_stat].values
        idx_k = np.logical_and(x_s > x_e.min(), x_s < x_e.max())
        thetas_inside_obs_bounds = np.where(idx_k)[0]
        if len(thetas_inside_obs_bounds) == 0:
            xmax = t_s.max()
            xmin = t_s.min()
        else:
            xmax = np.quantile(t_s[thetas_inside_obs_bounds], .95)
            xmin = np.quantile(t_s[thetas_inside_obs_bounds], .05)

        fig, ax = plt.subplots(1,1)
        ax.scatter(t_s, x_s, alpha=.1)
        ax.set_ylim(0, x_e.max()*4)
        ax.set_xlim(xmin, xmax)
        ax.set_ylabel(stats_names[n_summary_stat])
        ax.set_xlabel(prior.keys[n_theta].split(".")[-1])
        ax.set_xscale(xscale)
        ax.hlines([x_e.min(), x_e.max()], [xmin], [xmax], color="tab:orange")
        f = f"{xscale}x__{prior.keys[n_theta].split('.')[-1]}__{stats_names[n_summary_stat]}"
        d = os.path.join(plot_output, "prior_predictive_checks")
        os.makedirs(d, exist_ok=True)
        fig.savefig(os.path.join(d, f))
        
    N, K = x_sim.shape
    idx = np.repeat(True, N)
    fig, axes = plt.subplots(K, 1, figsize=(8,8))
    
    for k in range(K):
        x_e = x_experiment[:, k].values
        x_s = x_sim[:, k]

        idx_k = np.logical_and(x_s > x_e.min(), x_s < x_e.max())
        idx = np.logical_and(idx, idx_k)
        
        axes[k].hist(x_s, color="tab:blue")
        ax = axes[k].twinx()
        ax.hist(x_e, color="tab:orange")
        ax.plot(
            np.nan, 
            label=f"{stats_names[k]}, overlap {idx_k.sum()}/{N}", 
            linestyle=""
        )
        ax.legend()

    # calculate the amount of simulation results which are inside the 
    # experimental observations
    n_inside_bounds = idx.sum()

    fig.suptitle(
        f"Number of simulated results inside observed bounds: {n_inside_bounds} " +
        "(complete match)\n"+
        f"see {mod}.stats for the interpretation of the summary statistics"
    )

    if n_inside_bounds < 0.1 * N:
        axes[0].set_title(
            f"WARNING! Simulations inside bounds should be > 10% of sims ({N})",
            color="red"
        )

    fig.savefig(os.path.join(plot_output, "prior_predictive_checks.png"), dpi=300)
    
    Kt = theta_sim.shape[1]
    fig2, axes = plt.subplots(nrows=K, ncols=Kt, figsize=(8,12), sharex="col", sharey="row")
    fig2.text(0.5, 0.04, 'parameter', ha='center')
    fig2.text(0.04, 0.5, 'summary statistic', va='center', rotation='vertical')
    fig2.suptitle(
        "prior ~ simulated summary statistic\n"+
        "the windows should be more or less evenly filled."
    )
    
    for i in range(Kt):
        for j in range(K):
            plot_good_param_space(i, j, xscale="linear")
            plot_good_param_space(i, j, xscale="log")
            y = x_sim[:, j]
            x = theta_sim[:, i]
            axes[j,i].scatter(x, y, alpha=.1)
            axes[j,i].set_xlim(*np.quantile(x[~np.isnan(x)], (0.5,.95)))
            axes[j,i].set_ylim(*np.quantile(y[~np.isnan(y)], (0.5,.95)))
            axes[j,0].set_ylabel(stats_names[j])

        axes[K-1, i].set_xlabel(prior.keys[i].split(".")[-1])

    fig2.savefig(os.path.join(
        plot_output, 
        "prior_predictive_checks_prior_vs_result.png"), dpi=300)



if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
import numpy as np

class ErrorModel(ABC):
    @abstractmethod
    def forward(self, Y):
        """Obtain a realization of Y that depends on the chosen error function"""
        pass

    @abstractmethod
    def reverse(self, Y, Y_obs):
        """Obtain an error estimate of the difference between Y and Y_obs
        This difference depends on the chosen error function.
        """
        pass

    def __call__(self, Y, Y_obs=None):
        if Y_obs is None:
            return self.forward(Y=Y)
        else:
            return self.reverse(Y=Y, Y_obs=Y_obs)
        

class SumSquaredDifferenceErrorModel(ErrorModel):
    """This error model assumes that the error between Y and Y_obs is normal distributed
    It is one of the simplest error models
    """
    def forward(self, Y):
        return Y
    
    def reverse(self, Y, Y_obs):
        return np.sum(np.square(Y - Y_obs))


from functools import partial
from typing import Dict, Tuple, Union

import numpy as np
from numpy.random import Generator, PCG64
from scipy.stats._distn_infrastructure import rv_continuous, rv_discrete, rv_generic, rv_continuous_frozen, rv_discrete_frozen

from pymob.sim.parameters import scipy_to_scipy
from pymob.inference.base import Distribution, Errorfunction, InferenceBackend
from pymob.simulation import SimulationBase
from pymob.utils.config import lookup
from pymob.inference.error_models import ErrorModel

class ScipyDistribution(Distribution):
    """Distribution wrapper for SciPy random variables.

    This subclass of :class:`~pymob.inference.base.Distribution` provides a thin
    wrapper around SciPy's continuous and discrete distributions. It maps a
    distribution name to the corresponding SciPy ``rv_continuous`` or
    ``rv_discrete`` object using the ``scipy_to_scipy`` dictionary. The class
    also defines a ``parameter_converter`` that converts parameter arrays to
    NumPy ``ndarray`` objects.

    The primary purpose of this class is to expose a ``dist_name`` property that
    returns the name of the underlying SciPy distribution.
    """
    distribution_map: Dict[str,Tuple[Union[rv_continuous,rv_discrete],Dict[str,str]]] = scipy_to_scipy
    parameter_converter = staticmethod(lambda x: np.array(x))
    
    @property
    def dist_name(self) -> str:
        return self.distribution.name
    

    
class ScipyBackend(InferenceBackend):
    """Backend that uses SciPy distributions for inference.

    The backend implements the abstract :class:`~pymob.inference.base.InferenceBackend`
    interface using SciPy probability distributions. It parses the model priors
    and error models from the simulation configuration, builds a
    :class:`ProbabilisticModel` that can generate prior predictive samples,
    compute likelihoods, and sample from the prior distribution.

    Attributes
    ----------
    inference_model : ProbabilisticModel
        The assembled probabilistic model used for all inference operations.
    random_state : numpy.random.Generator
        Random number generator seeded from the simulation configuration.
    """
    _distribution = ScipyDistribution
    distribution: Union[rv_continuous,rv_discrete]

    def __init__(self, simulation: SimulationBase) -> None:
        super().__init__(simulation)
        self.inference_model = self.parse_probabilistic_model()
        self.random_state = Generator(PCG64(self.config.simulation.seed))

    def parse_deterministic_model(self):
        pass

    def parse_probabilistic_model(self):
        return ProbabilisticModel(
            prior_model=self.prior,
            error_model=self.error_model,
            indices=self.indices,
            observations=self.simulation.observations,
            simulation=self.simulation,
            eps=self.config.inference.eps,
            seed=self.config.simulation.seed,
        )


    def posterior_predictions(self):
        pass

    def prior_predictions(self):
        return self.inference_model(theta=None, observations=None)


    def sample_distribution(self):
        return self.inference_model.prior_model()
    
    def create_log_likelihood(self) -> Tuple[Errorfunction, Errorfunction]:
        """Create log-likelihood and (optional) gradient functions.

        Returns
        -------
        tuple
            A pair ``(log_likelihood, gradient)`` where each element is a
            callable conforming to the :class:`Errorfunction` protocol.
            The current backend does not implement a gradient, so both callables
            simply return ``None``. This stub satisfies the type checker and can be
            extended in the future.
        """
        # Simple placeholder implementations
        def _log_likelihood(theta):
            # Placeholder: actual implementation not provided
            return None

        def _gradient(theta):
            # Placeholder: gradient not implemented
            return None

        return _log_likelihood, _gradient


class ScipyPriorModel:
    """Helper class for sampling from prior distributions and computing log-probabilities.

    Instances maintain a reference to the prior model definition, the index
    variables, and observations. The ``__call__`` method forwards to either
    ``forward`` (when no parameters are supplied) or ``reverse`` (when a
    parameter dictionary is provided), enabling a uniform callable interface.

    Parameters
    ----------
    prior_model : dict
        Mapping of parameter names to :class:`Distribution` objects.
    indices : dict
        Index arrays for each indexed dimension of the simulation.
    observations : xarray.Dataset
        Observed data used for conditioning (currently not used in the prior).
    seed : int
        Seed for the underlying NumPy random generator.
    """
    def __init__(self, prior_model, indices, observations, seed):
        self.prior_model = prior_model
        self.random_state = Generator(PCG64(seed))
        
        self.context = [
            indices, observations
        ]

    def __call__(self, theta=None):
        if theta is None:
            return self.forward()
        else:
            return self.reverse(theta=theta)

    def _draw_random_variables(self) -> Dict[str, rv_generic]:
        prior_samples = {}

        # prior is added here, so it is updated 
        context = [prior_samples] + self.context
        for name, prior in self.prior_model.items():

            dist = prior.construct(context=context)


            sample = dist.rvs(size=prior.shape, random_state=self.random_state)

            prior_samples.update({name:sample})

        return prior_samples


    def _calc_log_prob(self, theta) -> Dict[str, rv_generic]:
        prior_samples = {}

        # prior is added here, so it is updated 
        context = [prior_samples] + self.context
        for name, prior in self.prior_model.items():

            dist = prior.construct(context=context)

            if hasattr(dist, "logpdf"):
                logprob = dist.logpdf(theta[name])
            elif hasattr(dist, "logpmf"):
                logprob = dist.logpmf(theta[name])
            else:
                raise NotImplementedError(
                    "scipy distribution must by rv_continuous or rv_discrete"
                )

            prior_samples.update({name: logprob})

        return prior_samples

    def forward(self):
        return self._draw_random_variables()
    
    def reverse(self, theta):
        return self._calc_log_prob(theta=theta)


class ScipyErrorModel(ErrorModel):
    """Error model that generates observation noise using SciPy distributions.

    The class builds a set of random variables based on the user-specified error
    model expressions. It can draw synthetic noisy observations from the model
    (``forward``) or compute the log-probability of observed data given a set of
    latent variables (``reverse``).

    Parameters
    ----------
    eps : float
        Small constant added to scales to avoid division by zero.
    error_model : dict
        Mapping of data variable names to error model :class:`Distribution` objects.
    indices : dict
        Index arrays for the simulation.
    observations : xarray.Dataset
        Observed data used for likelihood evaluation.
    seed : int
        Seed for the random number generator.
    """
    def __init__(self, eps, error_model, indices, observations, seed):
        extra = {"EPS": eps, "np": np}
        self.error_model = error_model
        self.random_state = Generator(PCG64(seed))

        self.context = [
            indices,
            observations,
            extra
        ]


    def _parameterize_random_variables(self, Y) -> Dict[str, rv_generic]:
        """Parameterizes random variables from Expression-based error models
        """
        distributions = {}
        for error_model_name, error_model_dist in self.error_model.items():
            rv = error_model_dist.construct(
                context=[Y] + self.context
            )

            distributions.update({error_model_name: rv})
                                
        return distributions


    def forward(self, Y):
        random_variables = self._parameterize_random_variables(Y=Y)
        return {
            key: rv.rvs(random_state=self.random_state) 
            for key, rv in random_variables.items()
        }

    def reverse(self, Y, Y_obs):
        random_variables = self._parameterize_random_variables(Y=Y)

        likelihoods = {}
        for key, rv in random_variables.items():
            if hasattr(rv, "logpdf"):
                logprob = rv.logpdf(Y_obs[key])
            elif hasattr(rv, "logpmf"):
                logprob = rv.logpmf(Y_obs[key])
            else:
                raise NotImplementedError(
                    "scipy distribution must by rv_continuous or rv_discrete"
                )
            
            likelihoods.update({key: logprob})

        return likelihoods
    

class ScipyTransModel:
    """Transformation model that runs the deterministic simulation.

    This lightweight wrapper forwards the ``theta`` (parameter values) to the
    simulation's ``dispatch`` method, executes the model, and returns the
    resulting simulated state ``Y``. It is used by the probabilistic model to
    generate latent system states from sampled parameters.
    """
    def __init__(self, simulation):
        self.simulation = simulation

    def transform_prior_to_error_model(self, theta, y0={}, x_in={}, seed=None):
        evaluator = self.simulation.dispatch(theta=theta, y0=y0, x_in=x_in)
        evaluator(seed)
        return evaluator.Y
    
    def __call__(self, theta, y0={}, x_in={}, seed=None):
        return self.transform_prior_to_error_model(theta, y0, x_in, seed)


class ProbabilisticModel:
    """Combined prior, transformation, and error model for inference.

    The class orchestrates three components:

    * **Prior model** - draws parameter samples and evaluates their log-probability.
    * **Transformation model** - runs the deterministic simulation to obtain latent
      system states.
    * **Error model** - generates synthetic observations or evaluates the likelihood
      of observed data.

    The ``__call__`` method implements four usage modes (prior predictive,
    likelihood, sampling, posterior predictive) as described in the docstring of
    ``ScipyBackend``.
    """

    def __init__(
        self,
        prior_model,
        error_model,
        indices,
        observations,
        simulation,
        eps,
        seed
    ):
        rng = np.random.default_rng(seed=seed)
        # split the seed into two additional random seeds
        seeds = rng.integers(low=1, high=1000, size=2)

        self.prior_model = ScipyPriorModel(
            prior_model=prior_model, indices=indices, observations=observations, seed=seeds[0]
        )
        self.trans_model = ScipyTransModel(
            simulation=simulation
        )
        self.error_model = ScipyErrorModel(
            eps=eps, error_model=error_model, indices=indices, observations=observations, seed=seeds[1]
        )

    def __call__(self, theta=None, observations=None):
        """Evaluate the inference model in various modes and calculate likelihoods

        prior predictions
        -----------------
        - theta = None
        - observations = None 

        When no paramaters and observations are specified, this corresponds to prior predictions
        parameters samples and noisy observations are generated
        
        
        likelihood
        ----------
        - theta defined
        - observations defined

        No random draws are made, but the probabilites of the passed parameters and 
        observations are computed

        
        sampling
        --------
        - theta = None
        - observations defined

        This reduces to a rudimentary markov sampler that generates random draws from
        the parameter distributions and evaluates the given observations w.r.t to
        the error model

        
        posterior_predictive
        --------------------
        - theta defined
        - observations = None

        When theta is defined and observations are undefined, random draws of the observations
        are generated, based on the error distributions

        
        """
        if theta is None:
            theta = self.prior_model()

        
        theta_prob = self.prior_model(theta)
        results = self.trans_model(theta)

        if observations is None:
            observations = self.error_model(Y=results)
        
        error_prob = self.error_model(Y=results, Y_obs=observations)

        return {
            "theta": theta, 
            "theta_prob": theta_prob, 
            "results": results,
            "observations": observations, 
            "observations_prob": error_prob
        }



from functools import partial, lru_cache
import glob
import re
import warnings
from typing import (
    Tuple, Dict, Union, Optional, Callable, Literal, List, Any,
    Protocol
)
import importlib

from tqdm import tqdm
import numpy as np
import xarray as xr
import arviz as az
from matplotlib import pyplot as plt
import sympy
from arviz.data.inference_data import SUPPORTED_GROUPS_ALL

import pymob
from pymob.simulation import SimulationBase
from pymob.sim.parameters import Expression, NumericArray
from pymob.inference.base import Errorfunction, InferenceBackend, Distribution
from pymob.inference.analysis import (
    cluster_chains, rename_extra_dims, plot_posterior_samples,
    add_cluster_coordinates
)
from pymob.inference.numpyro_dist_map import (
    scipy_to_numpyro, transformed_dist_map, transform, inv_transform
)

import numpyro
from numpyro.infer import Predictive
from numpyro.distributions import Normal
from numpyro.distributions.distribution import DistributionMeta
from numpyro import handlers
from numpyro import infer

import jax
import jax.numpy as jnp
import sympy2jax

SUPPORTED_GROUPS_ALL.extend([
    "prior_model_fits", 
    "prior_residuals",
    "posterior_model_fits",
    "posterior_residuals",
    "unconstrained_prior",
    "unconstrained_posterior"
])


sympy2jax_extra_funcs = {
    sympy.Array: jnp.array,
    sympy.Tuple: tuple,
}


def catch_patterns(expression_str):
    # tries to match array notation [0 1 2]
    pattern = r"\[(\d+(\.\d+)?(\s+\d+(\.\d+)?)*|\s*)\]"
    if re.fullmatch(pattern, expression_str) is not None:
        expression_str = expression_str.replace(" ", ",") \
            .removeprefix("[").removesuffix("]")
        return f"stack({expression_str})"

    return expression_str

# distribution_map =  {
#     "lognorm": (LogNormalTrans, {"scale": "loc", "s": "scale"}),
#     "binom": (dist.Binomial, {"n":"total_count", "p":"probs"}),
#     "normal": dist.Normal,
#     "halfnorm": HalfNormalTrans,
#     "poisson": (dist.Poisson, {"mu": "rate"}),
# }

class ErrorModelFunction(Protocol):
    def __call__(
        self, 
        theta: Dict, 
        simulation_results: Dict, 
        indices: Dict,
        observations: Dict, 
        masks: Dict,
        make_predictions: bool
    ) -> Any:
        ...

class NumpyroDistribution(Distribution):
    distribution_map: Dict[str,Tuple[DistributionMeta, Dict[str,str]]] = scipy_to_numpyro
    parameter_converter = staticmethod(lambda x: jnp.array(x))

    def _get_distribution(self, distribution: str) -> Tuple[DistributionMeta, Dict[str, str]]:
        # TODO: This is not satisfying. I think the transformed distributions
        # should only be used when this is explicitly specified.
        # I really wonder, why this makes such a large change in numpyro
        numpyro_dist, parameter_mapping = self.distribution_map[distribution]
        if numpyro_dist in transformed_dist_map:
            transformed_dist = transformed_dist_map[numpyro_dist]
            return transformed_dist, parameter_mapping
        else:
            # could not find the transformed distribution using normal
            return numpyro_dist, parameter_mapping

    @property
    def dist_name(self):
        return self.distribution.__name__

class NumpyroBackend(InferenceBackend):
    _distribution = NumpyroDistribution
    prior: Dict[str, DistributionMeta]

    def __init__(
        self, 
        simulation: SimulationBase
    ):
        """Initializes the NumpyroBackend with a Simulation object.

        Parameters
        ----------
        simulation : SimulationBase
            An initialized simulation.
        """
        super().__init__(simulation=simulation)
        # parse preprocessing
        if self.user_defined_preprocessing is not None:
            self.preprocessing = getattr(
                self.simulation._prob,
                self.user_defined_preprocessing
            )

        # parse the probability model
        if self.user_defined_probability_model is not None:
            inference_model = getattr(
                self.simulation._prob, 
                self.user_defined_probability_model
            )
            self.inference_model = partial(
                inference_model,
                indices = self.indices
            )

        # combine the model
        else:
            self.inference_model = self.parse_probabilistic_model()

        if self.user_defined_error_model is not None:
            if "." in self.user_defined_error_model:
                # if "." in the name of the model assume that this is the 
                # fully quailified name of the function (including module)
                module, func = self.user_defined_error_model.rsplit(".", 1)
                
                # import the module
                _module = importlib.import_module(module)
            else:
                # if not qualified name
                # use plain name, assuming this is the function
                func = self.user_defined_error_model

                # use already imported _prob module from the case study
                _module = self.simulation._prob

            user_error_model = getattr(_module, func)

            self.inference_model = partial(
                self.inference_model,
                user_error_model=user_error_model
            )

        self.check_tolerance_and_jax_mode()
    

    @property
    def user_defined_probability_model(self):
        return self.config.inference_numpyro.user_defined_probability_model
    
    @property
    def user_defined_error_model(self):
        return self.config.inference_numpyro.user_defined_error_model
    
    @property
    def user_defined_preprocessing(self):
        return self.config.inference_numpyro.user_defined_preprocessing

    @property
    def gaussian_base_distribution(self):
        return self.config.inference_numpyro.gaussian_base_distribution
    
    @property
    def chains(self):
        if self.config.inference_numpyro.kernel == "svi":
            return 1
        else:
            return self.config.inference_numpyro.chains
    
    @property
    def draws(self):
        return self.config.inference_numpyro.draws
    
    @property
    def warmup(self):
        return self.config.inference_numpyro.warmup
    
    @property
    def thinning(self):
        return self.config.inference_numpyro.thinning

    @property
    def svi_iterations(self):
        return self.config.inference_numpyro.svi_iterations
    
    @property
    def svi_learning_rate(self):
        return self.config.inference_numpyro.svi_learning_rate
    
    @property
    def kernel(self):
        return self.config.inference_numpyro.kernel
    
    @property
    def adapt_state_size(self):
        return self.config.inference_numpyro.sa_adapt_state_size

    @property
    def init_strategy(self):
        strategy = self.config.inference_numpyro.init_strategy
        return getattr(infer, strategy)

    @staticmethod
    def generate_transform(expression: Expression):
        # check for parentheses in expression
        
        # Parse the expression without knowing the symbol names in advance
        parsed_expression = sympy.sympify(str(expression), evaluate=False)
        free_symbols = tuple(parsed_expression.free_symbols)

        # Transform expression to jax expression
        func = sympy2jax.SymbolicModule(
            parsed_expression, 
            extra_funcs=sympy2jax_extra_funcs, 
            make_array=True
        )

        return {"transform": func, "args": [str(s) for s in free_symbols]}

    def check_tolerance_and_jax_mode(self):
        x64 = jax.config.read("jax_enable_x64")
        atol = self.config.jaxsolver.atol
        if atol < 1e-8 and not x64:
            warnings.warn(
                "Jax is not running in 64 bit mode but the precision is smaller "+
                f"than 1e-8 (config.jaxsolver.atol={atol} < 1e-8). "
                "Increase the absolute tolerance value or run jax in 64 bit mode. "
                "Script: `jax.config.update('jax_enable_x64', True)`. "
                "Commandline: Set environmental variable `JAX_ENABLE_X64=True`.",
                category=UserWarning
            )
        else:
            print("Jax 64 bit mode:", x64)
            print("Absolute tolerance:", atol)


    def parse_deterministic_model(self) -> Callable:
        """Parses an evaluation function from the Simulation object, which 
        takes a single argument theta and defaults to passing no seed to the
        deterministic evaluator.

        Returns
        -------
        callable
            The evaluation function
        """
        def evaluator(theta, y0={}, x_in={}, seed=None):
            evaluator = self.simulation.dispatch(theta=theta, y0=y0, x_in=x_in)
            evaluator(seed)
            return evaluator.Y
        
        return evaluator

    def model(self):
        pass


    def observation_parser(self) -> Tuple[Dict,Dict]:
        """Transform a xarray.Dataset into a dictionary of jnp.Arrays. Creates
        boolean arrays of masks for nan values (missing values are tagged False)

        Returns
        -------
        Tuple[Dict,Dict]
            Dictionaries of observations (data) and masks (missing values)
        """
        obs = self.simulation.observations #\
            # .transpose(*self.simulation.dimensions)
        y0 = self.simulation.model_parameters.get("y0", {})
        observed_data_vars = self.config.data_structure.observed_data_variables + self.extra_vars

        if len(observed_data_vars) == 0:
            warnings.warn(
                "No observed data_variables were found. Is this correct? "+
                "Make sure you have marked "+
                "All relevant data variables as observed before running inference"+
                "`sim.config.data_variables.MY_DATA_VAR.observed = True`",
                category=UserWarning
            )

        # this was implemented to deal with import export and the required precision
        # for jax 
        if jax.config.jax_enable_x64:
            bit = "64"
        else:
            bit = "32"

        masks = {}
        observations = {}
        for d in observed_data_vars:
            o = jnp.array(self.cast_to_precision(obs[d].values, precision=bit))
            m = jnp.logical_not(jnp.isnan(o))
            observations.update({d:o})
            masks.update({d:m})
        
        # add y0 values to the observation dict
        for d in self.config.data_structure.data_variables:
            if d in y0:
                observations.update({
                    f"{d}_y0": jnp.array(self.cast_to_precision(y0[d], precision=bit))
                })

        return observations, masks
    
    @staticmethod
    def cast_to_precision(value, precision="64"):
        if precision in str(value.dtype):
            return value
        
        if np.issubdtype(value.dtype, np.floating):
            return value.astype(f"float{precision}")
        elif np.issubdtype(value.dtype, np.integer):
            return value.astype(f"int{precision}")
        else:
            return value

    def parse_probabilistic_model(self):
        EPS = self.EPS
        prior = self.prior.copy()
        # TODO: This should be passed to the model, becuase if the batch
        #       dimension is subsetted, the index size is not anymore the same
        #       and correspondingly, be parsed during parse_obs or so
        indices = self.indices
        error_model = self.error_model.copy()
        extra = {"EPS": EPS, "jnp": jnp, "np": np}
        gaussian_base = self.gaussian_base_distribution
        data_variables_y0 = [
            f"{dv}_y0" for dv in self.config.data_structure.data_variables
        ]

        def sample_prior(prior: Dict, obs: Dict, indices: Dict):
            theta = {}
            context = [theta, indices, obs, extra]
            for prior_name, prior_dist in prior.items():
                if prior_dist._dist_str == "deterministic":
                    theta_i = prior_dist.construct(
                        context=context, 
                        extra_kwargs={"name": prior_name}
                    )
                else:  
                    dist = prior_dist.construct(context=context)

                    # TODO: distribution expansion is not so trivial unfortunately
                    dist = dist.expand(batch_shape=prior_dist.shape)

                    theta_i = numpyro.sample(
                        name=prior_name,
                        fn=dist,
                    )
                theta.update({prior_name: theta_i})

            return {}, theta
        
        def sample_prior_gaussian_base(prior: Dict, obs: Dict, indices=indices):
            theta = {}
            theta_base = {}
            context = [theta, indices, obs, extra]
            for prior_name, prior_dist in prior.items():
                if prior_dist._dist_str == "deterministic":
                    theta_i = prior_dist.construct(
                        context=context, 
                        extra_kwargs={"name": prior_name}
                    )
                else:    
                    dist = prior_dist.construct(context=context)

                    try:
                        transforms = getattr(dist, "transforms")
                    except:
                        raise RuntimeError(
                            f"The specified distribution {prior_dist} had no transforms. If setting "+
                            "the option 'inference.numpyro.gaussian_base_distribution = 1', "+
                            "you are only allowed to use parameter distribution, which can "+
                            "be specified as transformed normal distributions. "+
                            f"Currently {transformed_dist_map.keys()} are specified"+
                            "You can use the numypro.distributions.TransformedDistribution "+
                            "API to specify additional distributions with transforms."+
                            "And pass them to the inferer by updating the distribution map: "+
                            "sim.inferer.distribution_map.update({'newdist': your_new_distribution})"
                        )

                    # sample from a random normal distribution
                    # CHECK: Expanding before 
                    # TODO: potentially just use the base dist here, because this is then
                    # already truncated
                    theta_base_i = numpyro.sample(
                        name=f"{prior_name}_normal_base",
                        fn=dist.base_dist.expand(batch_shape=prior_dist.shape),
                    )


                    # apply the transforms 
                    theta_i = numpyro.deterministic(
                        name=prior_name,
                        value=transform(transforms=transforms, x=theta_base_i)
                    )

                    theta_base.update({prior_name: theta_base_i})
                theta.update({prior_name: theta_i})

            return theta_base, theta

        def likelihood(theta, simulation_results, indices, observations, masks, make_predictions):
            """Uses lookup and error model from the local function context"""
            context = [simulation_results, theta, indices, observations, extra]
            for error_model_name, error_model_dist in error_model.items():
                dist = error_model_dist.construct(context=context)

                # this assumes that the transform function is only ever 
                # characterized by algebraic transforms of the observations 
                # and the determinsitic model of that variable 
                # This should cover the vast majority of cases.
                if error_model_dist.obs_transform_func is None:
                    # TODO: consider if this should go into the 2nd ifelse condition
                    #       because then, I could actually rename _obs and _res 
                    #       and downstream use this information for creating inverse
                    #       function
                    if make_predictions:
                        obs = None
                    else:
                        obs = observations[error_model_name]
                    
                    _ = numpyro.sample(
                        name=error_model_name + "_obs",
                        fn=dist.mask(masks[error_model_name]),
                        obs=obs
                    )
                else:
                    residuals = numpyro.deterministic(
                        name=error_model_name + "_res",
                        value=error_model_dist.obs_transform_func(
                            obs=observations[error_model_name], 
                            **{error_model_name: simulation_results[error_model_name]}, 
                        )
                    )

                    if make_predictions:
                        residuals = None

                    _ = numpyro.sample(
                        name=error_model_name + "_obs",
                        fn=dist.mask(masks[error_model_name]),
                        obs=residuals
                    )


        def model(
            solver, obs, masks, 
            only_prior: bool = False, 
            user_error_model: Optional[ErrorModelFunction] = None,
            make_predictions: bool = False,
        ):
            # construct priors with numpyro.sample and sample during inference
            if gaussian_base:
                theta_gaussian, theta_ = sample_prior_gaussian_base(
                    prior=prior, 
                    obs=obs, 
                    indices=indices,
                )
            else:
                _, theta_ = sample_prior(
                    prior=prior, 
                    obs=obs, 
                    indices=indices,
                )
            
            if only_prior:
                return
            
            # calculate deterministic simulation with parameter samples
            y0 = {k.replace("_y0",""): v for k, v in theta_.items() if k in data_variables_y0}
            theta = {k: v for k, v in theta_.items() if k not in data_variables_y0}
            sim_results = solver(theta=theta, y0=y0)

            # store data_variables as deterministic model output
            for deterministic_name, deterministic_value in sim_results.items():
                _ = numpyro.deterministic(
                    name=deterministic_name, 
                    value=deterministic_value
                )

            if user_error_model is None:
                _ = likelihood(
                    theta=theta_,
                    simulation_results=sim_results,
                    indices=indices,
                    observations=obs,
                    masks=masks,
                    make_predictions=make_predictions,
                )
            else:
                _ = user_error_model(
                    theta=theta,
                    simulation_results=sim_results,
                    indices=indices,
                    observations=obs,
                    masks=masks,
                    make_predictions=make_predictions,
                )

        return model

    @staticmethod
    def preprocessing(**kwargs):
        return kwargs

    def run(self, print_debug=True, render_model=True):
        # set parameters of JAX and numpyro
        # jax.config.update("jax_enable_x64", True)

        # generate random keys
        key = jax.random.PRNGKey(self.simulation.config.simulation.seed)
        key, *subkeys = jax.random.split(key, 20)
        keys = iter(subkeys)

        # parse observations and masks for missing data
        obs, masks = self.observation_parser()

        model_kwargs = self.preprocessing(
            obs=obs, 
            masks=masks,
        )

        # prepare model and print information about shapes
        model = partial(
            self.inference_model, 
            solver=self.evaluator, 
            **model_kwargs
        )    

        if render_model:
            try:
                import graphviz
                graph = numpyro.render_model(model, render_distributions=False)
                graph.render(
                    filename=f"{self.simulation.output_path}/probability_model",
                    view=False, cleanup=True, format="png",
                ) 
            except graphviz.backend.ExecutableNotFound:
                warnings.warn(
                    "Model is not rendered, because the graphviz executable is "+
                    "not found. Try search for 'graphviz executables not found' "+
                    "and the used OS. This should be an easy fix :-)"
                )


        if print_debug:
            with numpyro.handlers.seed(rng_seed=1):
                trace = numpyro.handlers.trace(model).get_trace()
            print(numpyro.util.format_shapes(trace)) # type: ignore
            
        # run inference
        if self.kernel.lower() == "sa" or self.kernel.lower() == "nuts":
            sampler, mcmc = self.run_mcmc(
                model=model,
                keys=keys,
                kernel=self.kernel.lower()
            )

            # create arviz idata
            self.idata = self.nuts_posterior(
                mcmc=mcmc, model=model, key=next(keys), obs=obs
            )

        elif self.kernel.lower() == "svi" or self.kernel.lower() == "map":
            if not self.gaussian_base_distribution:
                raise RuntimeError(
                    "SVI is only supported if parameter distributions can be "+
                    "re-parameterized as gaussians. Please set "+
                    "inference.numpyro.gaussian_base_distribution = 1 "+
                    "and if needed use distributions from the loc-scale family "+
                    "to specify the model parameters."
                )
            
            svi, guide, svi_result = self.run_svi(
                model=model,
                keys=keys,
                kernel=self.kernel.lower(),
                learning_rate=self.svi_learning_rate,
                iterations=self.svi_iterations,
            )

            # plot loss curve
            fig, ax = plt.subplots(1, 1)
            ax.plot(svi_result.losses)
            ax.set_yscale("log")
            ax.set_ylabel("Loss")
            ax.set_xlabel("Iteration")
            fig.savefig(f"{self.simulation.output_path}/svi_loss_curve.png")

            # save idata and print summary
            draws = 1 if self.kernel.lower() == "map" else self.draws
            self.idata = self.svi_posterior(
                svi_result, model, guide, next(keys), draws
            )
            print(az.summary(self.idata))

        else:
            raise NotImplementedError(
                f"Kernel {self.kernel} is not implemented. "+
                "Use one of nuts, sa, svi, map"
            )
        
    def run_mcmc(self, model, keys, kernel):
        if kernel == "sa":
            sampler = infer.SA(
                model=model,
                dense_mass=True,
                adapt_state_size=self.config.inference_numpyro.sa_adapt_state_size,
                init_strategy=self.init_strategy,
            )

        elif kernel == "nuts":
            sampler = infer.NUTS(
                model, 
                dense_mass=self.config.inference_numpyro.nuts_dense_mass, 
                step_size=self.config.inference_numpyro.nuts_step_size,
                adapt_mass_matrix=self.config.inference_numpyro.nuts_adapt_mass_matrix,
                adapt_step_size=self.config.inference_numpyro.nuts_adapt_step_size,
                max_tree_depth=self.config.inference_numpyro.nuts_max_tree_depth,
                target_accept_prob=self.config.inference_numpyro.nuts_target_accept_prob,
                init_strategy=self.init_strategy,
            )
        else:
            raise NotImplementedError(
                f"MCMC kernel {kernel} not implemented. Use one of 'sa', 'nuts'"
            )


        mcmc = infer.MCMC(
            sampler=sampler,
            num_warmup=self.warmup,
            num_samples=self.draws * self.thinning,
            num_chains=self.chains,
            thinning=self.thinning,
            progress_bar=True,
        )
    
        # run inference
        mcmc.run(next(keys))
        mcmc.print_summary()

        return sampler, mcmc

    @staticmethod
    def run_svi(model, keys, learning_rate, iterations, kernel):

        init_fn = partial(infer.init_to_uniform, radius=1)
        if kernel == "svi":
            guide = infer.autoguide.AutoMultivariateNormal(model, init_loc_fn=init_fn)
        elif kernel == "map":
            guide = numpyro.infer.autoguide.AutoDelta(model, init_loc_fn=init_fn)
        else:
            raise NotImplementedError(
                f"SVI kernel {kernel} is not implemented. "+
                "Use one of 'map', 'svi'"
            )

        optimizer = numpyro.optim.ClippedAdam(step_size=learning_rate, clip_norm=10)
        svi = infer.SVI(model=model, guide=guide, optim=optimizer, loss=infer.Trace_ELBO())
        svi_result = svi.run(next(keys), iterations, stable_update=True)

        if kernel == "svi":
            cov = svi_result.params['auto_scale_tril'].dot(
                svi_result.params['auto_scale_tril'].T
            )
            median = guide.median(svi_result.params)

        return svi, guide, svi_result

    @property
    def posterior(self):
        warnings.warn(
            "Discouraged use of inferer.posterior API"+
            "use inferer.idata.posterior instead."
        )
        return self.idata.posterior  # type: ignore

    def create_log_likelihood(
        self, 
        seed=1,
        return_type:Literal["joint-log-likelihood", "full", "summed-by-site", "summed-by-prior-data", "custom"]="joint-log-likelihood",
        check=True,
        custom_return_fn: Optional[Callable] = None,
        scaled=True,
        vectorize=False,
        gradients=False,
    ) -> Tuple[Errorfunction,ErrorModelFunction]:
        """Log density relies heavily on the substitute utility
        
        The log density is the scaled log-likelihood. In case the the scale handler
        is  used, log_density reflects this. Usually, the scaled log-density
        should be returned, because it is loss used for the optimizer/sampler

        The general method is actually quite simple. Values of all SAMPLE
        sites are replaced according to the key: value pairs in `theta`.

        Then the model is calculated and the trace is obtained. Everything
        else is then just post-processing of the sites. Here the log_prob
        function of the sites in the trace are used and the values of the
        sites are inserted. 

        Note that the log-density can randomly fluctuate, if not all
        sites are replaced.

        Note that the data-loglik can be used to calculate a maximum-likelihood
        estimate. Because it is independent of the prior

        The method is equivalent using the log_likelihood method, but returns
        only the likelihood of the data given the model parameters.
        
        .. :code-block: python
        
           def log_likelihood(theta: dict):
               theta = {f"{key}_normal_base": val for key, val in theta.items()}
               loglik = numpyro.infer.util.log_likelihood(
                   model=seeded_model,
                   posterior_samples=theta, 
                   batch_ndims=0,
                   solver=sim.inferer.evaluator,
                   obs=data,
                   masks=masks,
               )
              return loglik
           
           jax.vmap(log_likelihood)(theta)

        Parameters
        ----------


        return_type : str
            The information which should be returned. With increasing level
            of computation:
            
            joint-log-likelihood: returns a single value, the entire log
                likelihood of the model, given the values in theta
            full: joint-log, loglik-prior of each site and value, 
                loglik-data of each site and value 
            summed-by-site: joint-loglik, loglik-prior of sites, 
                loglik-data of sites
            summed-by-prior-data:
                joint-loglik, prior-loglik, data-loglik
            custom:
                uses the full log

        """
        key = jax.random.PRNGKey(seed)
        obs, masks = self.observation_parser()

        model_kwargs = self.preprocessing(
            obs=obs, 
            masks=masks,
        )
        
        # prepare model
        model = partial(
            self.inference_model, 
            solver=self.evaluator, 
            **model_kwargs
        )    

        seeded_model = numpyro.handlers.seed(model, key)
   
        def log_density(
            theta: Dict[str, float|List[NumericArray]], 
        ):
            """
            Calculate the log-probability of different sites of the probabilistic
            model

            Parameters
            ----------

            theta : Dict
                Dictionary of priors (sites) which should be deterministically 
                fixed (substituted).

            """
            
            if self.gaussian_base_distribution:
                theta_ = theta.copy()
                theta = {}
                for key, val in theta_.items():
                    if "_normal_base" in key:
                        theta_i = {key: val}
                    else:
                        theta_i = {f"{key}_normal_base": val}

                    theta.update(theta_i)
            else:
                pass
        
            joint_log_density, trace = numpyro.infer.util.log_density( # type: ignore
                model=seeded_model,
                model_args=(),
                model_kwargs={},
                params=theta
            )

            joint_log_density = joint_log_density

            if check:
                sites_in_theta = {
                    name: True 
                    if name in theta
                    else False
                    for name, site in {
                        name: site for name, site in trace.items() 
                        if site["type"] == "sample" 
                        and not site["is_observed"]
                    }.items()
                }

                if not all(list(sites_in_theta.values())):
                    missing_sites = [name for name, in_theta in sites_in_theta.items() if not in_theta]
                    warnings.warn(
                        f"Sites: {missing_sites} were not specified in theta. "+
                        "Log-likelihood will not be fully defined by theta. "+
                        "Results should be independent of the given seed"
                    )


            if return_type == "joint-log-likelihood":
                return joint_log_density
            
            def get_scale(site):
                # get the scale factor for the log probability
                scale = site["scale"]
                if scale is None or not scaled:
                    return jnp.array(1.0, dtype=float)
                else:
                    return scale
            
            prior_loglik = {
                name: site["fn"].log_prob(site["value"]) * get_scale(site)
                for name, site in trace.items()
                if site["type"] == "sample" and not site["is_observed"]
            }

            data_loglik = {
                name: site["fn"].log_prob(site["value"]) * get_scale(site)
                for name, site in trace.items()
                if site["type"] == "sample" and site["is_observed"]
            }

            if return_type == "full":
                return joint_log_density, prior_loglik, data_loglik


            if return_type == "custom":
                return custom_return_fn(joint_log_density, prior_loglik, data_loglik)

            prior_loglik_sum = {
                key: jnp.sum(value) for key, value in prior_loglik.items()
            }

            data_loglik_sum = {
                key: jnp.sum(value) for key, value in data_loglik.items()
            }

            if return_type == "summed-by-site":
                return joint_log_density, prior_loglik_sum, data_loglik_sum

            prior_loglik_total = jnp.sum(jnp.array(list(prior_loglik_sum.values())))
            data_loglik_total = jnp.sum(jnp.array(list(data_loglik_sum.values())))
            
            if return_type == "summed-by-prior-data":
                return joint_log_density, prior_loglik_total, data_loglik_total
            
            raise NotImplementedError(f"return_type flag: {return_type} is not implemented")

        if gradients:
            if not (return_type == "joint-log-likelihood" or return_type == "custom"):
                raise ValueError(
                    "Gradients need a single return value to be computed. "+
                    f"Either choose `return_type={'joint-log-likelihood'}` or "+
                    f"`return_type={'custom'}` and specify an approxpriate "+
                    "`custom_return_fn`."
                )
            else:
                grad_log_density = jax.grad(log_density)

        else:
            grad_log_density = lambda x: None

        if vectorize:
            return jax.vmap(log_density), jax.vmap(grad_log_density)

        else:
            return log_density, grad_log_density

    def check_log_likelihood(
        self, 
        theta: Optional[Dict[str, float|NumericArray]]=None,
        vectorize=False
    ):
        log_density, _ = self.create_log_likelihood(
            seed=self.config.simulation.seed,
            return_type="full",
            check=True,
            vectorize=vectorize
        )
        
        if theta is not None:
            pass
        elif self.config.inference_numpyro.gaussian_base_distribution:
            theta = {k: 0.0 for k, _ in self.config.model_parameters.free.items()}
        else:
            # TODO: replace by prior sample, or prior mean, ...
            theta = self.config.model_parameters.value_dict
            
        
        llsum, llpri, lldat = log_density(theta=theta)
        if not np.isnan(llsum) and not np.isnan(llsum):
            return llsum, llpri, lldat
        
        nanlogliks_pri = [k for k, g in llpri.items() if np.any(np.isnan(g)) or np.any(np.isinf(g))]
        nanlogliks_dat = [k for k, g in lldat.items() if np.any(np.isnan(g)) or np.any(np.isinf(g))]
        
        if len(nanlogliks_dat + nanlogliks_pri) > 0:
            warnings.warn(
                f"Log-likelihoods {nanlogliks_dat + nanlogliks_pri} contained "+
                "NaN or inf values. The gradient based "+
                "samplers will not be able to sample from this model. Make sure "+
                "that all functions are numerically well behaved. "+
                "Inspect the model with `jax.debug.print('{}',x)` "+
                "https://jax.readthedocs.io/en/latest/notebooks/external_callbacks.html#exploring-debug-callback "+
                "Or look at the functions step by step to find the position where "+
                "jnp.grad(func)(x) evaluates to NaN"
            )
        return llsum, llpri, lldat

    def check_gradients(
        self, 
        theta: Optional[Dict[str, float|NumericArray]]=None,
        vectorize=False
    ):

        _, grad_log_density = self.create_log_likelihood(
            seed=self.config.simulation.seed,
            return_type="joint-log-likelihood",
            check=False,
            gradients=True,
            vectorize=vectorize
        )

        if theta is not None:
            pass
        elif self.config.inference_numpyro.gaussian_base_distribution:
            # this works, because in the grad function the parameters are
            # renamed with f'{k}_normal_base' if the gaussian base distribution
            # is used.
            theta = {k: 0.0 for k, _ in self.config.model_parameters.free.items()}
        else:
            # TODO: replace by prior sample, or prior mean, ...
            theta = self.config.model_parameters.value_dict
            
        grads = grad_log_density(theta)
        nangrads = [k for k, g in grads.items() if np.any(np.isnan(g)) or np.any(np.isinf(g))]
        if len(nangrads) > 0:
            warnings.warn(
                f"Gradients {nangrads} contained NaN or infinity values. The gradient based "+
                "samplers will not be able to sample from this model. Make sure "+
                "that all functions are numerically well behaved. "+
                "Inspect the model with `jax.debug.print('{}',x)` "+
                "https://jax.readthedocs.io/en/latest/notebooks/external_callbacks.html#exploring-debug-callback "+
                "Or look at the functions step by step to find the position where "+
                "jnp.grad(func)(x) evaluates to NaN"
            )

        return grads


    def predict_observations(self, model, posterior_samples, key, n=100):
        """
        there is a very small remark in the numpyro API that explains that
        if data input for observed variables is None, the data are sampled
        from the distributions instead of returning the input data
        https://num.pyro.ai/en/stable/getting_started.html#a-simple-example-8-schools
        """
        data_vars = self.config.data_structure.observed_data_variables
        predictive = Predictive(
            model=partial(model, make_predictions=True),
            # use all samples that were not generated in the likelihood function
            posterior_samples={
                k: v for k, v in posterior_samples.items() if k not in 
                [f"{k}_obs" for k in data_vars] + [f"{k}_res" for k in data_vars]
            },
            return_sites=[f"{k}_obs" for k in data_vars],
            num_samples=n, 
            batch_ndims=2
        )
        # using the same key, to generate predictions the same way
        # as for the the posterior samples
        predictions = predictive(key)

        if self.user_defined_probability_model is not None:
            return predictions

        observations = {}
        for data_var in self.config.data_structure.observed_data_variables:
            if data_var not in self.error_model:
                # catches the case where a user defined error model was provided
                # and error models were not defined.
                # TODO: Use trace handler to obtain the different components
                #       prior, likelihood and deterministic transforms to populate
                #       missing components
                if f"{data_var}_obs" not in predictions:
                    raise KeyError(
                        f"`{data_var}` was not found in the predictions, "+
                        "but is marked as an observed data variable. Either "+
                        f"set `sim.config.data_variables.{data_var}.observed = False` " +
                        f"or generate predictions for '{data_var}' in the error-model."
                    )
                obs = predictions[f"{data_var}_obs"]
            else:
                if self.error_model[data_var].obs_transform_func is not None:
                    transform_inv = self.error_model[data_var].obs_transform_func_inv
                    if transform_inv is None:
                        warnings.warn(
                            "Cannot make predictions of observations from normalized "+
                            "observations (residuals). Please provide an inverse observation "+
                            f"transform: e.g. `sim.config.error_model['{data_var}'].obs_inv = ...`."+
                            "residuals are denoted as 'res'. See Lotka-volterra case "+
                            "study for an example. "
                        )
                        # this will fetch the residuals (they are called _obs)
                        obs = predictions[f"{data_var}_obs"]
                    else:
                        obs = transform_inv(
                            res=predictions[f"{data_var}_obs"],
                            **{data_var: posterior_samples[data_var]}
                        )
                    
                    # get the deterministic residuals res=y_obs-y_det regardless
                    res = posterior_samples[f"{data_var}_res"] 
                    observations.update({f"{data_var}_res": res})
                else:
                    obs = predictions[f"{data_var}_obs"]

            observations.update({f"{data_var}_obs": obs})

        return observations

    def calculate_log_likelihood(self, model, posterior_samples):
        return numpyro.infer.log_likelihood(
            model=model, 
            posterior_samples=posterior_samples, 
            batch_ndims=2, 
        )



    @lru_cache
    def prior_predictions(self, n=None, seed=1):
        if n is None:
            n = self.n_predictions
            
        key = jax.random.PRNGKey(seed)
        obs, masks = self.observation_parser()

        model_kwargs = self.preprocessing(obs=obs, masks=masks)

        # prepare model
        model = partial(
            self.inference_model, 
            solver=self.evaluator, 
            **model_kwargs
        )    
   
        prior_predictive = Predictive(
            model, num_samples=n, batch_ndims=2
        )
        prior_samples = prior_predictive(key)

        log_likelihood = self.calculate_log_likelihood(
            model=model, posterior_samples=prior_samples
        )

        obs_predictions = self.predict_observations(
            model=model, posterior_samples=prior_samples, key=key, n=n
        )

        return self.to_arviz_idata(
            prior=prior_samples,
            prior_predictive=obs_predictions,
            observed_data=obs,
            log_likelihood=log_likelihood,
            n_draws=n,
            n_chains=1
        )
    

    def to_arviz_idata(
        self,
        prior: Dict[str, NumericArray] = {},
        posterior: Dict[str, NumericArray] = {},
        log_likelihood: Dict[str, NumericArray] = {},
        prior_predictive: Dict[str, NumericArray] = {},
        posterior_predictive: Dict[str, NumericArray] = {},
        observed_data: Dict[str, NumericArray] = {},
        n_draws: Optional[int] = None,
        n_chains: Optional[int] = None,
        **kwargs,
    ):
        """Create an Arviz idata object from samples.
        TODO: Outsource to base.InferenceBackend
        """
        posterior_coords = self.posterior_coordinates
        # posterior_coords["chain"] = [0]
        if n_draws is not None:
            posterior_coords["draw"] = list(range(n_draws))

        if n_chains is not None:
            posterior_coords["chain"] = list(range(n_chains))
        data_structure = self.posterior_data_structure

        data_variables = self.config.data_structure.data_variables
        obs_data_variables = self.config.data_structure.observed_data_variables
        prior_keys = self.simulation.model_parameter_dict.keys()
        
        prior_ = {k: v for k, v in prior.items() if k in prior_keys}
        unconstrained_prior_ = {
            k: v for k, v in prior.items() 
            if k in [f"{k}_normal_base" for k in prior_keys]
        }
        prior_model_fits_ = {
            k: v for k, v in prior.items() if k in data_variables
        }
        prior_residuals_ = {
            k.replace("_res", ""): v 
            for k, v in prior_predictive.items() 
            if k in [f"{d}_res" for d in data_variables]
        }
        prior_predictive_ = {
            k.replace("_obs", ""): v 
            for k, v in prior_predictive.items() 
            if k in [f"{d}_obs" for d in data_variables]
        }

        posterior_ = {k: v for k, v in posterior.items() if k in prior_keys}
        unconstrained_posterior_ = {
            k: v for k, v in posterior.items() 
            if k in [f"{k}_normal_base" for k in prior_keys]
        }
        posterior_model_fits_ = {
            k: v for k, v in posterior.items() if k in data_variables
        }
        posterior_residuals_ = {
            k.replace("_res", ""): v 
            for k, v in posterior_predictive.items() 
            if k in [f"{d}_res" for d in data_variables]
        }
        posterior_predictive_ = {
            k.replace("_obs", ""): v 
            for k, v in posterior_predictive.items() 
            if k in [f"{d}_obs" for d in data_variables]
        }
        
        likelihood_ = {
            k.replace("_obs", ""): v 
            for k, v in log_likelihood.items() 
            if k in [f"{d}_obs" for d in obs_data_variables]
        }
        
        observed_data_ = {
            k: v for k, v in observed_data.items() 
            if k in obs_data_variables
        }
        
        if len(prior_) > 0 and len(prior_keys) != len(prior_):
            miss_keys = [k for k in prior_keys if k not in prior]
        else:
            miss_keys = []

        if len(posterior_) > 0 and len(prior_keys) != len(posterior_):
            miss_keys = [k for k in prior_keys if k not in posterior]
        else:
            miss_keys = []

        if len(miss_keys) > 0:
            warnings.warn(
                f"Parameters {miss_keys} "+
                "were not found in the predictions. Make sure that the prior "+
                "names match the names in user_defined_probability_model = "+
                f"{self.config.inference_numpyro.user_defined_probability_model}.",
                category=UserWarning
            )

        idata = az.from_dict(
            prior=prior_,
            prior_predictive=prior_predictive_,
            posterior=posterior_,
            posterior_predictive=posterior_predictive_,
            observed_data=observed_data_,
            log_likelihood=likelihood_,
            **kwargs,
            dims=data_structure,
            coords=posterior_coords,
        )

        idata.add_groups(group_dict={
                "prior_model_fits": prior_model_fits_,
                "posterior_model_fits": posterior_model_fits_,
                "prior_residuals": prior_residuals_,
                "posterior_residuals": posterior_residuals_,
                "unconstrained_prior": unconstrained_prior_,
                "unconstrained_posterior": unconstrained_posterior_,
            }, 
            dims=data_structure,
            coords=posterior_coords
        )

        for group_key in idata.groups():
            if self.config.simulation.batch_dimension in idata[group_key].coords:
                idata[group_key] = idata[group_key].assign_coords({  # type: ignore
                  k: (self.config.simulation.batch_dimension, v) 
                    for k, v in self.indices.items()
                }) 
                idata[group_key].attrs.update({"pymob_version": pymob.__version__})

        # this automatically adds clusters for multiple chains
        # by default this is done by testing if the means of all parameters
        # deviate by more than one standard deviation.
        # The method does not take any decisions, it just labels the chains
        # with a specific cluster.
        idata = add_cluster_coordinates(idata, deviation="std")

        return idata
    

    def posterior_draws_from_svi(self, guide, svi_result, n, key):
        # prepare model without obs, so obs are sampled from the posterior
        params = svi_result.params

        # this gets the parameters of the normal_base_distributions
        predictive = Predictive(
            guide, params=params, 
            num_samples=n, batch_ndims=2
        )
        posterior_samples = predictive(key)
        return posterior_samples


    def svi_posterior(self, svi_result, model, guide, key, n=1000):
        # TODO: Harmonize SVI posterior and nuts posterior. The base function
        # should only accept only samples from the posterior. Then the tool
        # could be used to generate observations with higher time resolution
        obs, masks = self.observation_parser()

        posterior_samples = self.posterior_draws_from_svi(
            guide=guide, svi_result=svi_result, n=n, key=key
        )

        # this gets all the parameters and predictions
        predictive = Predictive(
            model, posterior_samples, #params=params, 
            num_samples=n, batch_ndims=2
        )
        posterior = predictive(key)

        # passing complete posterior is identical to only passing only posterior_samples
        log_likelihood = self.calculate_log_likelihood(
            model=model, posterior_samples=posterior_samples, 
        )

        # here only the transformed data are passed. I suppose this is enough,
        # because. The transformed data are the only thing required for the
        obs_predictions = self.predict_observations(
            model=model, posterior_samples=posterior, key=key, n=n
        )

        # TODO: When passing the complete posterior to the calculate_log_likelihood
        # function the results are the same as when only using the posterior,
        # but when passing it to predict_observations, the results are different
        # I suppose this is correct, but make sure to udnerstand what's going on
        # and then document it.
        complete_posterior = {}
        complete_posterior.update(posterior_samples)
        complete_posterior.update(posterior)

        return self.to_arviz_idata(
            posterior=complete_posterior,
            posterior_predictive=obs_predictions,
            log_likelihood=log_likelihood,
            observed_data=obs,
            n_draws=n
        )



    def nuts_posterior(self, mcmc, model, key, obs):
        samples = jax.device_get(mcmc.get_samples(group_by_chain=True))

        priors = list(self.prior.keys())
        data_variables = self.config.data_structure.data_variables
        
        log_likelihood = self.calculate_log_likelihood(
            model=model, posterior_samples=samples
        )
        
        posterior_predictive = self.predict_observations(
            model, posterior_samples=samples, key=key, n=None
        )

        unconstrained_prior = [f"{p}_normal_base" for p in priors]
        return self.to_arviz_idata(
            posterior={
                k: v for k, v in samples.items() 
                if k in priors + unconstrained_prior + data_variables
            },
            posterior_predictive=posterior_predictive,
            log_likelihood=log_likelihood,
            observed_data=obs,
            sample_stats=mcmc.get_extra_fields(group_by_chain=True)
        )

    @staticmethod
    def get_dict(group: xr.Dataset):
        data_dict = group.to_dict()["data_vars"]
        return {k: np.array(val["data"]) for k, val in data_dict.items()}


    @lru_cache
    def posterior_predictions(self, n: Optional[int]=None, seed=1):
        # TODO: It may be necessary that the coordinates should be passed as 
        # constant data. Because if the model is compiled with them once, 
        # according to the design philosophy of JAX, the model will not 
        # be evaluated again. But considering that the jitted functions do take
        # coordinates as an input argument, maybe I'm okay. This should be
        # tested.
        posterior = self.idata.posterior  # type: ignore
        n_samples = posterior.sizes["chain"] * posterior.sizes["draw"]
        data_variables_y0 = [
            f"{dv}_y0" for dv in self.config.data_structure.data_variables
        ]

        if n is not None:
            key = jax.random.PRNGKey(seed)

            # use the minimum of the desired and available samples
            n = min(n, n_samples)

            n_draws = int(n / posterior.sizes["chain"])
            # the same selection of draws will be applied to all chains. This
            # any other form will result in an array, where a lot of nans 
            # are present of the size n * chains, while we want size n
            selection = jax.random.choice(
                key=key, 
                a=posterior.draw.values, 
                replace=False, 
                shape=(n_draws, ) # type: ignore
            )
            posterior = posterior.isel(draw=selection)


        preds = []
        with tqdm(
            total=posterior.sizes["chain"] * posterior.sizes["draw"],
            desc="Posterior predictions"
        ) as pbar:
            for chain in posterior.chain:
                for draw in posterior.draw:
                    theta_arr = posterior.sel(draw=draw, chain=chain)
                    theta_dict = self.get_dict(theta_arr)
                    
                    # calculate deterministic simulation with parameter samples
                    y0 = {k.replace("_y0",""): v for k, v in theta_dict.items() if k in data_variables_y0}
                    theta = {k: v for k, v in theta_dict.items() if k not in data_variables_y0}
                    evaluator = self.simulation.dispatch(theta=theta, y0=y0)
                    evaluator()
                    ds = evaluator.results

                    ds = ds.assign_coords({"chain": chain, "draw": draw})
                    ds = ds.expand_dims(("chain", "draw")) # type:ignore
                    preds.append(ds)
                    pbar.update(1)

        # key = jax.random.PRNGKey(seed)
        # model = partial(self.model, solver=self.evaluator)    
        # predict = numpyro.infer.Predictive(model, posterior_samples=posterior, batch_ndims=2)
        # predict(key, obs=obs, masks=masks)
        

        return xr.combine_by_coords(preds)

    def store_results(self, output=None):
        if output is not None:
            self.idata.to_netcdf(output)
        else:
            self.idata.to_netcdf(f"{self.simulation.output_path}/numpyro_posterior.nc")

    def load_results(self, file="numpyro_posterior.nc", cluster: Optional[int] = None):
        idata = az.from_netcdf(f"{self.simulation.output_path}/{file}")
        if cluster is not None:
            self.select_cluster(idata, cluster)

        self.idata = idata

    @staticmethod
    def select_cluster(idata: az.InferenceData, cluster: int):
        for g in idata.groups():
            idata_group = idata[g].where(idata[g].cluster == cluster, drop=True)
            del idata[g]
            idata[g] = idata_group

        warnings.warn(
            f"FILTERING POSTERIOR: Selecting cluster {cluster} from the complete "
            f"posterior. The cluster contains the chains: {idata_group.chain.values}."
        )
        return idata

    # This is a separate script!    
    def combine_chains(self, chain_location="chains", drop_extra_vars=[], cluster_deviation="std"):
        """Combine chains if chains were computed in a fully parallelized manner
        (on different machines, jobs, etc.). 

        In addition, the method drops all data variables and '..._norm' priors 
        (i.e. helper priors with a normal base). This is done, in order to
        create slim data objects for storage.

        Parameters
        ----------
        chain_location : str, optional
            location of the chains, relative to the simulation.output_path, this
            parameter is simulteneously the string appended to the saved 
            posterior. By default "chains"
        drop_extra_vars : List, optional
            any additional variables to drop from the posterior
        """
        sim = self.simulation
        pseudo_chains = glob.glob(
            f"{sim.output_path}/{chain_location}/*/numpyro_posterior.nc"
        )

        # just be aware that in the case of MAP this is not an acutal posterior.
        # But it can behave like one with multiple chains (1 for each start)
        idata = az.from_netcdf(pseudo_chains[0])
        posterior = self.drop_vars_from_posterior(idata.posterior, drop_extra_vars) # type: ignore
        log_likelihood = idata.log_likelihood # type: ignore

        # iterate over the posterior files with a progress bar (depending on the
        # size and number of posteriors this op needs time and memory)
        tqdm_iterator = tqdm(
            enumerate(pseudo_chains[1:], start=1), 
            total=len(pseudo_chains)-1,
            desc="Concatenating posteriors"
        )
        for i, f in tqdm_iterator:
            idata = az.from_netcdf(f)
            ccord = {"chain": np.array([i])}
            
            # add chain coordinate to posterior and likelihood
            idata.posterior = self.drop_vars_from_posterior(idata.posterior, drop_extra_vars) # type: ignore
            idata.posterior = idata.posterior.assign_coords(ccord) # type: ignore
            idata.log_likelihood = idata.log_likelihood.assign_coords(ccord) # type: ignore

            # concatenate chains
            posterior = xr.concat([posterior, idata.posterior], dim="chain") # type: ignore
            log_likelihood = xr.concat(
                [log_likelihood, idata.log_likelihood], # type: ignore
                dim="chain"
            )

        posterior = rename_extra_dims(
            posterior, 
            new_dim="substance", 
            new_coords=sim.observations.attrs["substance"]
        )

        # store mutlichain inferencedata to the main output directory
        # this is also a slim posterior that only contains the necessary information
        # posterior and likelihood and has therefore a small file size
        idata_multichain = az.InferenceData(
            posterior=posterior, 
            log_likelihood=log_likelihood,
            observed_data=idata.observed_data, # type: ignore
        )

        idata_multichain = add_cluster_coordinates(idata_multichain, cluster_deviation)
        print("Clusters:", idata_multichain.posterior.cluster)

        return idata_multichain            

    def drop_vars_from_posterior(self, posterior, drop_extra_vars):
        """drops extra variables if they are included in the posterior
        """
        drop_vars = [k for k in list(posterior.data_vars.keys()) if "_norm" in k]
        drop_vars = drop_vars + self.config.data_structure.data_variables + drop_extra_vars
        drop_vars = [v for v in drop_vars if v in posterior]
        drop_coords = [c for c in list(posterior.coords.keys()) if c.split("_dim_")[0] in drop_vars]

        posterior = posterior.drop(drop_vars)
        posterior = posterior.drop(drop_coords)

        return posterior


import pytest

from tests.fixtures import init_simulation_casestudy_api


def test_interactive_mode():
    sim = init_simulation_casestudy_api()
    sim.interactive()


if __name__ == "__main__":
    import sys
    import os
    sys.path.append(os.getcwd())
    # test_scripting_API()
    # test_interactive_mode()



import pytest
import xarray as xr
import numpy as np
from click.testing import CliRunner

from pymob.simulation import SimulationBase
from pymob.sim.config import Param, DataVariable

from tests.fixtures import init_simulation_casestudy_api, linear_model

def test_simulation():
    sim = init_simulation_casestudy_api()

    evalu = sim.dispatch(theta=sim.model_parameter_dict)
    evalu()

    ds = evalu.results
    ds_ref = xr.load_dataset(f"{sim.data_path}/simulated_data.nc")

    np.testing.assert_allclose(
        (ds - ds_ref).to_array().values,
        0
    )

def test_minimal_simulation():
    sim = SimulationBase()
    linreg, x, y, y_noise, parameters = linear_model()

    obs = xr.DataArray(y_noise, coords={"x": x}).to_dataset(name="y")
    sim.observations = obs
    
    from pymob.sim.solvetools import solve_analytic_1d
    sim.model = linreg
    sim.solver = solve_analytic_1d

    sim.config.model_parameters.a = Param(value=10, free=False)
    sim.config.model_parameters.b = Param(value=3, free=True , prior="normal(loc=0,scale=10)") # type:ignore
    sim.model_parameters["parameters"] = sim.config.model_parameters.value_dict

    sim.dispatch_constructor()
    evaluator = sim.dispatch(theta={"b":3})
    evaluator()
    evaluator.results

    np.testing.assert_allclose(evaluator.results.y.values, y * 3 + 10)

    # this tests automatic updating of the parameterize method with partial
    sim.config.model_parameters.a = Param(value=0, free=False)
    sim.model_parameters["parameters"] = sim.config.model_parameters.value_dict
    evaluator = sim.dispatch(theta={"b":3})
    evaluator()
    evaluator.results

    np.testing.assert_allclose(evaluator.results.y.values, y * 3)

    sim.config.model_parameters.sigma_y = Param(free=True , prior="lognorm(scale=1,s=1)") # type:ignore
    sim.config.error_model.y = "normal(loc=y,scale=sigma_y)"

    sim.set_inferer("numpyro")
    sim.inferer.config.inference_numpyro.kernel = "nuts"
    # sim.inferer.config.inference_pyabc.min_eps_diff = 0.001
    sim.inferer.run()
    b = float(sim.inferer.idata.posterior["b"].mean()) # type: ignore
    sigma_y = float(sim.inferer.idata.posterior["sigma_y"].mean()) # type: ignore

    # test that the _model parameters of the Simulation remain unchanged. This is
    # achieved throgh deepcopying the dictionary on setting partial
    assert sim._model_parameters["parameters"]["b"] == 3
    np.testing.assert_allclose(b, parameters["b"], atol=0.05, rtol=0.05)
    np.testing.assert_allclose(sigma_y, parameters["sigma_y"], atol=0.05, rtol=0.05)

def test_input_parsing():
    sim = SimulationBase()
    sim.config.data_structure.A = DataVariable(dimensions=["x", "y"], observed=False)

    sim.config.simulation.y0 = ["X=0"]
    
    try:
        sim.parse_input(input="y0")
        threw_error = False
    except KeyError:
        threw_error = True

    assert threw_error        


    sim.config.data_structure.X = DataVariable(dimensions=["x"])
    test_coordinates = {"x": np.linspace(1, 10, 5), "y": np.array([1, 2])}
    sim.coordinates = test_coordinates
    
    y0 = sim.parse_input(input="y0", drop_dims=[])
    np.testing.assert_equal(y0.X.values, np.zeros(sim.dimension_sizes["x"]))

    y0 = sim.parse_input(input="y0", drop_dims=["x"])
    assert float(y0.X.values) == 0.0


    sim.config.data_structure.X = DataVariable(dimensions=["y", "x"])
    y0 = sim.parse_input(input="y0", drop_dims=[])
    np.testing.assert_equal(y0.X.values, np.zeros((2, 5)))


    sim.config.data_structure.X = DataVariable(dimensions=["x", "y"])
    y0 = sim.parse_input(input="y0", drop_dims=[])
    np.testing.assert_equal(y0.X.values, np.zeros((5, 2)))

    # test if broadcasting is done correctly
    sim.config.data_structure.X = DataVariable(dimensions=["x", "y"])
    sim.config.simulation.y0 = ["X=Array([0, 1])"]
    y0 = sim.parse_input(input="y0", drop_dims=[])
    np.testing.assert_equal(y0.X.isel(y=0), np.zeros((5)))
    np.testing.assert_equal(y0.X.isel(y=1), np.ones((5)))

    # test if broadcasting is done correctly
    sim.config.data_structure.X = DataVariable(dimensions=["y", "x"])
    sim.config.simulation.y0 = ["X=Array([0, 1, 2, 3,4])"]
    y0 = sim.parse_input(input="y0", drop_dims=[])
    np.testing.assert_equal(y0.X.isel(y=0), np.arange(5))
    np.testing.assert_equal(y0.X.isel(y=1), np.arange(5))

    # test if broadcasting is done correctly
    sim.config.data_structure.X = DataVariable(dimensions=["y", "x"])
    sim.config.simulation.y0 = ["X=Array([[0,1,2,3,4],[1,2,3,4,5]])"]
    y0 = sim.parse_input(input="y0", drop_dims=[])
    np.testing.assert_equal(y0.X.isel(y=0), np.arange(5))
    np.testing.assert_equal(y0.X.isel(y=1), np.arange(1, 6))

    # test if broadcasting is done correctly
    sim.config.data_structure.X = DataVariable(dimensions=["y", "x"])
    sim.config.simulation.y0 = ["X=Array([[0,1,2,3,4],[1,2,3,4,5]])"]
    y0 = sim.parse_input(input="y0", drop_dims=[])
    np.testing.assert_equal(y0.X.isel(y=0), np.arange(5))
    np.testing.assert_equal(y0.X.isel(y=1), np.arange(1, 6))

    # test if broadcasting is done correctly
    sim.config.data_structure.X = DataVariable(dimensions=["y", "x"])
    sim.config.simulation.y0 = ["X=Array([0, 1])"]
    y0 = sim.parse_input(input="y0", drop_dims=["x"])
    np.testing.assert_equal(y0.X.values, np.array([0,1]))

    sim.config.data_structure.X = DataVariable(dimensions=["y", "x"])
    sim.config.simulation.y0 = ["X=0"]
    y0 = sim.parse_input(input="y0", drop_dims=["y"])
    np.testing.assert_equal(y0.X.values, np.array([0,0,0,0,0]))

    sim.config.data_structure.X = DataVariable(dimensions=["y", "x"])
    sim.config.simulation.y0 = ["X=0"]
    y0 = sim.parse_input(input="y0", drop_dims=["x","y"])
    assert float(y0.X.values) == 0.0

    X = xr.DataArray(
        data=np.random.normal(size=(5,2)), 
        coords=sim.coordinates
    ).to_dataset(name="X")

    # test addition of observations when data variables had been specified
    # before
    sim.config.data_structure.X = DataVariable(dimensions=["y", "x"])
    sim.observations = X
    assert all([np.all(sc == tc) for (sk,sc), (tk,tc) in 
                zip(sim.coordinates.items(), test_coordinates.items())])

    sim.config.data_structure.remove("X")
    sim.observations = X

    sim.config.simulation.y0 = ["X=X"]
    y0 = sim.parse_input(input="y0", reference_data=X.isel(y=0), drop_dims=["y"])
    np.testing.assert_equal(y0.X.values, X.isel(y=0).X.values)
    
    sim.config.simulation.y0 = ["X=2*X"]
    y0 = sim.parse_input(input="y0", reference_data=X.isel(y=0), drop_dims=["y"])
    np.testing.assert_equal(y0.X.values, X.isel(y=0).X.values * 2)

    sim.config.simulation.y0 = ["X=exp(X)"]
    y0 = sim.parse_input(input="y0", reference_data=X.isel(y=0, x=0), drop_dims=["y", "x"])
    np.testing.assert_equal(y0.X.values, np.exp(X.isel(y=0, x=0).X.values))

def test_indexing_simulation():
    pytest.skip()

def test_no_error_from_repeated_setup():
    sim = init_simulation_casestudy_api()  # already executes setup
    sim.setup()


def test_commandline_api_simulate():
    from pymob.simulate import main
    runner = CliRunner()
    
    args = "--case_study=lotka_volterra_case_study "\
        "--scenario=test_scenario"
    result = runner.invoke(main, args.split(" "))

    if result.exception is not None:
        raise result.exception


if __name__ == "__main__":
    import sys
    import os
    sys.path.append(os.getcwd())



import numpy as np

rng = np.random.default_rng(seed=1)

def linear_model(n=50):
    def model(x, a, b):
        return a + x * b

    parameters = dict(
        a=0,
        b=1,
        sigma_y=1,
    )

    x = np.linspace(-5, 5, n)
    y = model(x=x, a=parameters["a"], b=parameters["b"])
    y_noise = rng.normal(loc=y, scale=parameters["sigma_y"])

    return model, x, y, y_noise, parameters


from pymob.utils.errors import import_optional_dependency

from . import analysis
from . import base
from . import scipy_backend
from . import error_models

pyabc = import_optional_dependency("pyabc", errors="ignore")
if pyabc is not None:
    from . import pyabc_backend

pymoo = import_optional_dependency("pymoo", errors="ignore")
if pymoo is not None:
    from . import pymoo_backend

numpyro = import_optional_dependency("numpyro", errors="ignore")
if numpyro is not None:
    from . import numpyro_backend
    from . import numpyro_dist_map


import os
import glob
import click
import numpy as np
import xarray as xr

# import own packages
from pymob.utils.store_file import (
    prepare_casestudy, import_package, opt, prepare_scenario)
from pymob.sim.simulation import update_parameters
from pymob.utils import help

@click.command()
@click.option("-c", "--case_study", type=str, default="core_daphnia", 
              help=help.case_study)
@click.option("-s", "--scenario", type=str, default="expo_control", 
              help=help.scenario)
@click.option("-n", "--number_simulations", type=int, default=None)
@click.option("-p", "--posterior_cluster", type=str, default=None)
def main(case_study, scenario, number_simulations, posterior_cluster):
    # parse sbi config file
    config = prepare_casestudy((case_study, scenario), "sbi.cfg")
    sim_conf = prepare_scenario(case_study, scenario)

    # load case study section settings (some parts of the section are populated 
    # automatically by the call to store.prepare_casestudy(...))
    output = config["case-study"].get("output")
    package = config["case-study"].get("package")
    print(f"forwarding output to {output}", flush=True)

    # import case study and load necessary objects
    mod = import_package(package_path=package)
    summary_stats = getattr(mod.stats, "raw")
    generator = getattr(mod.sim, config["sbi"].get("generator", fallback="generator"))

    # other options (override order is always a=command line, b=config file, c=default)
    n_pp = int(opt(a=number_simulations, b=config["sbi"].get("n_posterior_predictions"), c=100))
    pc = opt(a=posterior_cluster, b=config["sbi"].get("posterior_cluster"), c="*")
    obs_interval = config["sbi"].get("posterior_obs_interval", fallback="table")
    observations = config["sbi"].get("posterior_obs")

    # load specific or all clusters of posteriors (if chains have not converged)
    clusters = glob.glob(os.path.join(output, f"posterior_{pc}.nc"))
    print(clusters)
    # adapt simulation config file
    sim_conf["display_plots"] = False
    sim_conf["experiment"]["observation_interval"] = obs_interval
    sim_conf["experiment"]["observations"] = observations.split(" ")
    sim_conf["simulation"]["fail_silently"] = True

    # create necessary directories
    os.makedirs(os.path.join(output, "posterior_predictions"), exist_ok=True)

    for i, filename in enumerate(clusters):
        posterior = xr.load_dataarray(filename)
        
        # set up and run simulation
        print(len(posterior.sample))
        random_samples = np.random.randint(0, len(posterior.sample), n_pp)
        # out = generator(summary=summary_stats, config=sim_conf, init=True)
        for j in random_samples:
            sim_conf = update_parameters(
                sim_conf, posterior.isel(sample=j).values, 
                posterior.parameter.values)
            out = generator(
                summary=summary_stats, config=sim_conf, 
                init=False, return_xarray=True)

            out.to_netcdf(os.path.join(
                output, "posterior_predictions", f"pred_cluster_{i+1}_sample_{j}.nc"
            ))


if __name__ == "__main__":
    main()



import warnings
import importlib
from datetime import datetime
from pathlib import Path
import tempfile
import inspect
from frozendict import frozendict, FrozenOrderedDict
from dataclasses import dataclass, field
from typing import List, Dict, Callable, Tuple, Optional, TypedDict
from pymob.solvers.base import SolverBase, mappar
import sympy as sp
from sympy.printing.codeprinter import CodePrinter
from sympy.printing.numpy import NumPyPrinter, JaxPrinter
from sympy.printing.python import PythonPrinter
from sympy.printing.pycode import PythonCodePrinter
from pymob.utils.config import get_return_arguments, dX_dt2X

def recurse_variables(X, func):
    X_ = []
    for x in X:
        if isinstance(x, list|tuple):
            x_ = recurse_variables(x, func)
            X_.append(x_)
        else:
            X_.append(func(x))
    
    return X_


def flatten(x) -> List:
    if isinstance(x, list|tuple):
        return [a for i in x for a in flatten(i)]
    else:
        return [x]
    

def get_source_and_docs(func) -> Tuple[str,str]:
    complete_source = inspect.getsource(func)
    docstring = inspect.getdoc(func)
    if docstring is None:
        docstring = ""
    source = complete_source.replace(
        f'\n    """{docstring}"""', ''
    )
    return source, docstring

class SolutionDict(TypedDict):
    algebraic_solutions: Dict[str, sp.Eq]
    compiled_function: Callable

@dataclass(frozen=True)
class SymbolicODESolver(SolverBase):
    extra_attributes = [
        "output_path",
        "scenario_path",
        "do_compile",
        "compile_on_init",
        "module_name"
    ]

    def __post_init__(self, *args, **kwargs):
        super().__post_init__(*args, **kwargs)
        
        # this should happen at initialization time.
        if self.compile_on_init:
            funcs = self.compile(
                functions=self.compiler_recipe, 
                module_name=self.module_name,
                compile=self.do_compile
            )
            object.__setattr__(self, "compiled_functions", frozendict(funcs))

    compiled_functions: FrozenOrderedDict[str,SolutionDict] = frozendict({})
    output_path: str = tempfile.gettempdir()
    scenario_path: str = tempfile.gettempdir()
    do_compile: bool = True 
    compile_on_init: bool = True 
    module_name: str = "symbolic_solution"

    def define_symbols(self):
        """Define the necessary symbols solely based on the function"""
        thetanames = mappar(
            self.model, {}, 
            exclude=["t", "dt", "y", "x_in", "Y", "X"], 
            to="names"
        )
        ynames = [dX_dt2X(a) for a in get_return_arguments(self.model)]

        # define symbols for t, Y, Y_0 and theta
        t = sp.Symbol("t", positive=True, real=True)
        Y = {y: sp.Function(y, positive=True, real=True) for y in ynames}
        Y_0 = {
            f"{y}_0": sp.Symbol(f"{y}_0", positive=True, real=True) 
            for y in ynames
        }
        theta = {p: sp.Symbol(p, positive=True, real=True) for p in thetanames}
        
        symbols = (t, Y, Y_0, theta)

        return symbols
        
    @staticmethod
    def define_ode_system(rhs, t, Y, Y_0, theta) -> List:
        # get the symbolic ODEs by inserting symbols into the RHS
        dY_dt = rhs(t, recurse_variables(Y, func=lambda x: x(t)), *theta)

        # Define the system of ODEs as Equations with the derivative of the different 
        # state variables y(t) (e.g. B(t)/dt, D(t)/dt, ..) and the respective rhs of the
        # odes.
        ode_system = [
            sp.Eq(sp.Derivative(y(t), t), dy_dt) # type: ignore
            for y, dy_dt in zip(flatten(Y), flatten(dY_dt))
        ] 
        return ode_system

    @classmethod
    def solve_ode_system(cls, rhs, t, Y, Y_0, theta) -> Tuple[List,List]:
        # solve the ODE system. Here, ics could also be used. There are more arguments,
        # but a simple dsolve works. Before solving the ODEs are expanded to make it
        # easier for the solver.
        ode_system = cls.define_ode_system(rhs, t, Y, Y_0, theta)
        solution = sp.dsolve([ode.expand() for ode in ode_system])

        # convert solution to a list if the ODE system is only one equation
        if not isinstance(solution, list):
            solution = [solution]

        return solution, ode_system

    @staticmethod
    def compute_integration_constants_and_substitute(
        solution: List[sp.Eq], 
        initial_conditions: Dict,
        t: sp.Symbol,
        Y: List[sp.Function], 
        Y_0: List[sp.Symbol],
        theta: List[sp.Symbol]
    ):
        # Calculate the integration constants and insert into the equations step 
        # by step

        # dictionary of solved integration cosntants and the final solutions
        solved_ics = {}
        final_solutions = []

        # iterate over the integrated solutions of the ODEs
        for sol in solution:
            # substitute any integration constants that have been already evaluated
            sol_tmp = sol.subs(solved_ics) # type: ignore

            # identify remaining integration constants after substitution
            integration_constants = [
                s for s in sol_tmp.free_symbols 
                if s not in flatten(Y + theta + Y_0 + [t,])
            ]

            # go over each unsolved integration constant
            for ic in integration_constants:

                # substitute t=0 and the initial conditions into the equation. 
                # Substitution of the initial conditions, converts symbols defined 
                # as functions to constants
                sol_tmp_substituted = sol_tmp\
                    .subs({t:0})\
                    .subs(initial_conditions)\
                    .expand() # type:ignore
                
                # then solve the equation for the integration constant
                ic_t0 = sp.solve(sol_tmp_substituted, ic, dict=True)
                
                # make sure there is only one solution. Zero solutions, mean that 
                # the equation could not be solved, more than one solution signify, 
                # that there are multiple solutions, e.g due to x**2=4 when solving
                # for x then x=2 and x=-2
                assert len(ic_t0) == 1
                
                # add the integration constant to the dictionary for the following
                # solutions
                solved_ics.update(ic_t0[0])
                
                # substitute the integration constant
                sol_pre = sol_tmp.subs(solved_ics)

                # then expand and simplify. Expanding is essential, because 
                # otherwise simplify may not work as expected.
                # simplify is a bit dangerous, because it is not a deterministic
                # procedure
                sol_fin = sp.simplify(sol_pre.expand()) # type: ignore
                final_solutions.append(sol_fin)

        return final_solutions, solved_ics

    def solve_for_t0(
        self, 
        rhs: Callable, 
        t: sp.Symbol, 
        Y: List[sp.Function], 
        Y_0: List[sp.Symbol], 
        theta: List[sp.Symbol]
    ):
        solution, ode_system = self.solve_ode_system(rhs, t, Y, Y_0, theta)
        # define the initial or boundary conditions for the ODE system. This could also
        # be used in the solve, but it works better if the initial conditions are only
        # used when solving for the integration constants.
        initial_conditions={x(0):x_0 for x, x_0 in zip(flatten(Y),flatten(Y_0))} # type:ignore

        # compute the final solutions
        solutions, integration_constants = self.compute_integration_constants_and_substitute(
            solution=solution,
            initial_conditions=initial_conditions,
            Y=Y,
            Y_0=Y_0,
            theta=theta,
            t=t,
        )

        return solutions, integration_constants, ode_system

    @property
    def compiler_recipe(self):
        return {"F": self.compile_model}
    
    def compile(
        self, 
        functions: Dict[str,Callable], 
        module_name: str = "symbolic_solution",
        compile=False
    ):
        """Compiles an ODE model to source code or reads it from disk if 
        unchanged. Compilation, writing to disk and re-loading the module again
        has the fundamental advantage that the code can be debugged
        """
        python_module = ModulePythonCode()
        
        compiled_functions = {}
        if not compile:
            for func_name, func_compiler in functions.items():
                # try to load the module and function
                _, compiled_func = self.load_compiled_code(
                    module_name=module_name, func_name=func_name
                )        
            
                # store loaded results in dictionary
                results = {
                    "algebraic_solutions": None,
                    "compiled_function": compiled_func
                }
                compiled_functions.update({func_name: results})

        else:
            for func_name, func_compiler in functions.items():

                func_solutions, func_code = func_compiler(
                    func_name=func_name,
                    compiled_functions=compiled_functions
                )

                # add the function to the python module
                python_module.functions = tuple(
                    flatten([python_module.functions, func_code])
                )

                results = {
                    "algebraic_solutions": func_solutions,
                }


                # write the whole module to disk and load the functions
                # this is currently necessary, because definitions may depend
                # on one another.
                code_file = Path(self.output_path, f"{module_name}.py")
                with open(code_file, "w") as f:
                    f.writelines(str(python_module))

                _, compiled_func = self.load_compiled_code(
                    module_name=module_name, func_name=func_name
                )

                results.update({"compiled_function": compiled_func})

                compiled_functions.update({func_name: results})

        return compiled_functions

    def test_function_change(self, func, compiled_func):
        source, docs = get_source_and_docs(func)

        # otherwise see if there were changes in the source code 
        # and if not don't recompile
        python_code_c, docs_c = get_source_and_docs(compiled_func)

        if (
            docs_c.split("### SOURCE ###")[1].strip("\n") 
            == source.strip("\n")
        ):
            changed = False
        else:
            changed = True
        
        return changed

    def compile_model(self, func_name="F", compiled_functions={}):
        # get the ode arguments and names
        symbols = self.define_symbols()
        t, Y, Y_0, theta = symbols

        # solve system of differential equations algebraically
        solutions, integ_constants, ode_system = self.solve_for_t0(
            self.model,
            t=t,
            Y=list(Y.values()),
            Y_0=list(Y_0.values()),
            theta=list(theta.values())
        )

        # collect matrix terms
        solution_dict = {}
        idx = 0
        for i, (yname, ysyms) in enumerate(Y.items()):
            if isinstance(ysyms, list|tuple):
                neqs = len(ysyms)
            else:
                neqs = 1
            eqs = solutions[slice(idx, idx + neqs)]
            if neqs == 1:
                eqs = eqs[0]

            solution_dict.update({yname: eqs})
            idx += neqs

        docstring = self.generate_docstring(self.model)

        python_code = FunctionPythonCode(
            func_name=func_name,
            x="t",
            lhs_0=("Y_0", tuple(Y_0.keys())),
            lhs=tuple(Y.keys()),
            rhs=tuple(recurse_variables(solution_dict.values(), lambda x: x.rhs.expand())),
            theta=("θ", tuple(theta.keys())),
            expand_arguments=False,
            printer=CustomNumpyPrinter(),
            modules=("numpy",),
            docstring=docstring
        )

        str(python_code)

        tex = self.to_latex(solutions=solutions)
        code_file = Path(self.output_path, f"{func_name}.tex")
        with open(code_file, "w") as f:
            f.writelines(tex)

        return solution_dict, python_code

    def to_latex(self, solutions):
        latex = []
        for sol in solutions:

            if isinstance(sol.rhs, sp.Piecewise):
                rhs_expand = sp.Piecewise(*[
                    (sp.expand(part), sp.expand(cond)) 
                    for part, cond in sol.rhs.args # type: ignore
                ])
                eq = sp.Eq(sol.lhs, rhs_expand)
            else:
                eq = sol

            eq_tex = sp.latex(eq)

            latex.append(eq_tex)

        return latex

    @staticmethod
    def generate_docstring(func):
        source_with_comments, ode_docstring = get_source_and_docs(func)

        # create docstring
        python_time = f"Function compiled at: {datetime.now()}\n"
        python_ode_docstring = f"ODE DOCSTRING: {ode_docstring}\n"
        python_docstring = (
            '"""\n'
            +python_ode_docstring
            +python_time
            +'### SOURCE ###'
            +'\n'
            +'\n'.join(source_with_comments.splitlines())
            +'\n### SOURCE ###'
            +'\n"""\n'
        )

        return python_docstring
    
    def load_compiled_code(self, module_name, func_name):
        # Create a module spec from the file location
        code_file = Path(self.output_path, f"{module_name}.py")

        if code_file.exists():
            spec = importlib.util.spec_from_file_location( # type: ignore
                f"{module_name}.py", 
                code_file
            )

            # Create a module object from the spec
            module = importlib.util.module_from_spec(spec) # type: ignore
            spec.loader.exec_module(module)
        else:
            module = None

        if module is not None:
            compiled_function = getattr(module, func_name, None)
            if compiled_function is None:
                warnings.warn(
                    f"{func_name} was not found in {module}. "
                    "Try setting 'func_name' to one of: "
                    f"{[k for k, v in inspect.getmembers(module, inspect.isfunction)]}"
                )
                
        else:
            compiled_function = None

        return module, compiled_function
    
class PiecewiseSymbolicODESolver(SymbolicODESolver):
    @staticmethod
    def jump_func(funcnames="F t_jump"):
        # this is the master equation for solving until certain time t_jump and then
        # continuing the solve from there. This is generic. It just needs a
        # function to determine the jump location from the initial conditions

        t, θ, Y_0, ε = sp.symbols("t θ Y_0 ϵ", positive=True)
        F, t_jump = sp.symbols(funcnames, cls=sp.Function)
        

        F_master = sp.Piecewise(
            (
                F(t - t_jump(Y_0, θ) + ϵ, F(t_jump(Y_0, θ) - ϵ, Y_0, θ), θ), 
                ((0 < t_jump(Y_0, θ)) & (t_jump(Y_0, θ) < t))
            ), 
            (
                F(t, Y_0, θ), 
                True
            )
        )
        return F_master, (F, t_jump), (t, θ, Y_0, ε)

    def jump_solution(self, func_name, funcnames="F t_jump", compiled_functions={}):
        F_master, (F, t_jump), (t, θ, Y_0, ε) = self.jump_func(funcnames)
        # This can be converted into code, by writing the code for F to disc
        # and the code for t_jump and then then afterwards defining the master
        # equation.
        python_code = FunctionPythonCode(
            func_name=func_name,
            x="t",
            lhs_0=("Y_0", ("Y_0",)),
            theta=("θ", ("θ", "ε")),
            lhs=("Y",),
            rhs=(F_master,),
            expand_arguments=True,
            printer=CustomNumpyPrinter(),
            modules=("numpy",),
            docstring=""
        )

        tex = sp.latex(sp.Eq(sp.Symbol(func_name), F_master))
        code_file = Path(self.output_path, f"{func_name}.tex")
        with open(code_file, "w") as f:
            f.writelines(tex)

        return F_master, python_code
    
    def t_jump(self, solutions):
        raise NotImplementedError(
            "A Piecewise Symbolic solver needs to implement a function 't_jump'"
        )


class CustomNumpyPrinter(NumPyPrinter):
    def _print_Function(self, func):
        func_name = func.func.__name__
        args = ", ".join([self._print(arg) for arg in func.args]) # type: ignore
        return f"{func_name}({args})"
    
    def _print_LambertW(self, func):
        # Convert LambertW to scipy.special.lambertw
        arg = self._print(func.args[0]) # type: ignore
        return f"scipy.special.lambertw({arg})"

@dataclass
class FunctionPythonCode:
    func_name: str = "f"
    x: Optional[str] = None
    theta: Tuple[Optional[str],Tuple] = (None, ()) 
    lhs_0: Tuple[Optional[str],Tuple] = (None, ())
    lhs: Tuple = ()
    rhs: Tuple = ()
    expand_arguments: bool = True
    docstring: str = ""
    modules: Tuple[str, ...] = ()
    printer: CodePrinter = CustomNumpyPrinter()
    extra_indent: int = 0

    def __str__(self):
        source = (
            self._indent(self.signature, n=0)
            + self._indent(self.docstring)
            + self._indent(self.body)
            + self._indent(self.return_statement)
        )

        return self._indent(source, n=self.extra_indent, eol="\n")
    
    @staticmethod
    def _indent(expr, n=4, eol="\n"):
        if len(expr) == 0:
            return expr
        exprs = expr.split("\n")
        exprs = [" " * n + expr + eol for expr in exprs]
        return "".join(exprs)
    
    @property
    def signature(self):
        definition = f"def {self.func_name}"
        args = []

        if self.x is not None:
            args.append(self.x)
        if self.lhs_signature is not None:
            args.append(self.lhs_signature)
        if self.theta_signature is not None:
            args.append(self.theta_signature)
        
        args = ", ".join(args)
        return f"{definition}({args}):"
    
    @property
    def lhs_signature(self):
        if self.expand_arguments:
            if len(self.lhs_0[1]) == 0:
                return None
            return ", ".join(self.lhs_0[1])
        else:
            if len(self.lhs_0[1]) == 0:
                return None
            return self.lhs_0[0]

    @property
    def theta_signature(self):
        if self.expand_arguments:
            if len(self.theta[1]) == 0:
                return None
            return ", ".join(self.theta[1])
        else:
            if len(self.theta[1]) == 0:
                return None
            return self.theta[0]

    @property
    def body(self):
        funcbody = []
        if not self.expand_arguments:
            if len(self.lhs_0[1]) > 0:
                lhs_0 = f"{', '.join(self.lhs_0[1])} = {self.lhs_0[0]}"
                funcbody.append("# extract variable")
                funcbody.append(lhs_0)

        if not self.expand_arguments:
            if len(self.theta[1]) > 0:
                theta = f"{', '.join(self.theta[1])} = {self.theta[0]}"
                funcbody.append("# extract parameters (theta)")
                funcbody.append(theta)

        assert len(self.lhs) == len(self.rhs)
        if len(self.lhs) > 0:
            F = self.printer.doprint(
                expr=self.rhs, 
                assign_to=self.lhs
            )
            funcbody.append("# compute equation")
            funcbody.append(str(F))

        return "\n".join(funcbody)
    
    @property
    def return_statement(self):
        return f"return {', '.join(self.lhs)}"
    
@dataclass
class ModulePythonCode:
    functions: Tuple[FunctionPythonCode, ...] = ()

    def __str__(self):
        imports = "\n".join(self.import_statements)
        funcs = "\n\n\n".join(self.function_definitions)

        return f"{imports}\n\n\n{funcs}"

    @property
    def function_definitions(self):
        return [str(f) for f in self.functions]

    @property
    def import_statements(self) -> List[str]:
        imports = []
        for f in self.functions:
            for i in f.modules:
                imports.append(f"import {i}")
        return list(set(imports))

    def save_module(self, path):
        with open(path, "w") as f:
            f.writelines(str(self))



from functools import partial
from types import ModuleType
from collections import OrderedDict
from typing import Optional, List, Dict, Literal, Tuple, OrderedDict
from pymob.solvers.base import mappar, SolverBase
from frozendict import frozendict
from dataclasses import dataclass, field
import jax.numpy as jnp
from jax import Array
import jax
import diffrax
from diffrax.solver.base import _MetaAbstractSolver
from diffrax import (
    diffeqsolve, 
    Dopri5, 
    Tsit5,
    Kvaerno5,
    ODETerm, 
    SaveAt, 
    PIDController, 
    RecursiveCheckpointAdjoint,
    LinearInterpolation,
)

Mode = Literal['r', 'rb', 'w', 'wb']


@dataclass(frozen=True, eq=True)
class JaxSolver(SolverBase):
    """
    see https://jax.readthedocs.io/en/latest/faq.html#strategy-3-making-customclass-a-pytree
    to make thinks robust

    Parameters
    ----------

    throw_exceptions: bool
        Default is True. The JaxSolver will throw an exception if it runs into a problem
        with the step size and return infinity. This is done, because likelihood based inference
        algorithms can deal with infinity values and consider the tested parameter
        combination impossible. If used without caution, this can lead to severely
        biased parameter estimates.
    """

    diffrax_solver: _MetaAbstractSolver|str = field(default=Dopri5)
    rtol: float = 1e-6
    atol: float = 1e-7
    pcoeff: float = 0.0
    icoeff: float = 1.0
    dcoeff: float = 0.0
    max_steps: int = int(1e5)
    throw_exception: bool = True

    x_in_jumps: Tuple[float, ...] = field(init=False, repr=False)

    def __post_init__(self, *args, **kwargs):
        super().__post_init__(*args, **kwargs)

        # set the diffrax solver if it is a string instance
        dfx_solver = self.diffrax_solver
        if isinstance(dfx_solver, str):
            diffrax_solver = getattr(diffrax, dfx_solver)
            if not isinstance(diffrax_solver, _MetaAbstractSolver):
                raise TypeError(
                    f"Solver {diffrax_solver} must be {_MetaAbstractSolver}"
                )
            object.__setattr__(self, "diffrax_solver", diffrax_solver)

        x_in_jumps = self._get_x_in_jumps()
        object.__setattr__(self, "x_in_jumps", x_in_jumps)

        hash(self)

    def _get_x_in_jumps(self) -> Optional[Tuple[float, ...]]:
        """In case there are multiple interpolations present, this function
        ensures that all jumps are handled appropriately
        """
        xs_vars = []
        for _, data_var_coords in self.coordinates_input_vars["x_in"].items():
            x_var = data_var_coords[self.x_dim]

            xs_vars.append(jnp.array(x_var, dtype=float))

        if len(xs_vars) == 0:
            return None
        else:
            # unique returns a sorted and unique array, so it is perfect for 
            # the job
            x_in_jumps = jnp.unique(jnp.concatenate(xs_vars))

            # we need to cut off any jumps that are greater or equal to the
            # last observation, so that no infinities are returned for the
            # last x that is evaluated by the solver
            x_in_jumps_trimmed = x_in_jumps[x_in_jumps < self.x[-1]]
            # x_in[0][self.coordinates_input_vars["x_in"][self.x_dim] < self.x[-1]]
            
            return tuple(x_in_jumps_trimmed.tolist())



    @partial(jax.jit, static_argnames=["self"])
    def solve(self, parameters: Dict, y0:Dict={}, x_in:Dict={}):
        
        
        X_in = self.preprocess_x_in(x_in)
        x_in_flat = [x for xi in X_in for x in xi]

        Y_0 = self.preprocess_y_0(y0)

        ode_args, pp_args = self.preprocess_parameters(parameters)

        initialized_eval_func = partial(
            self.odesolve_splitargs,
            odestates = tuple(y0.keys()),
            n_odeargs=len(ode_args),
            n_ppargs=len(pp_args),
            n_xin=len(x_in_flat)
        )
        
        loop_eval = jax.vmap(
            initialized_eval_func, 
            in_axes=(
                *[0 for _ in range(self.n_ode_states)], 
                *[0 for _ in range(len(ode_args))],
                *[0 for _ in range(len(pp_args))],
                *[0 for _ in range(len(x_in_flat))], 
            )
        )
        result = loop_eval(*Y_0, *ode_args, *pp_args, *x_in_flat)

        # if self.batch_dimension not in self.coordinates:    
        # this is not yet stable, because it may remove extra dimensions
        # if there is a batch dimension of explicitly one specified

        # there is an extra dimension added if no batch dimension is present
        # this is added at the 0-axis
        # if parameters are scalars, the returned shape is 
        for v, val in result.items():
            if self.batch_dimension not in self.data_structure_and_dimensionality[v]:
                # otherwise it has a dummy dimension of length 1
                val_reduced = jnp.squeeze(val, 0)
            else:
                val_reduced = val

            expected_dims = tuple(self.data_structure_and_dimensionality[v].values())
            if len(expected_dims) != len(val_reduced.shape):
                # if the number of present dims is larger than the number of
                # expected dims, this is because the ODE "only" returned scalar
                # values. This is broadcasted to array of ndim=1
                val_reduced = jnp.squeeze(val_reduced, -1)
            else:
                pass

            # si = [
            #     s for dim, s in self.data_structure_and_dimensionality[v].items() 
            #     if dim != self.batch_dimension
            # ]
            
            # correct_shape = (s0, *si)
            
            # [i for i, vs in enumerate(val.shape) if vs not in expected_dims]
            # jnp.permute_dims(val, expected_dims)
            # val_reduced = val.permute_dims(expected_dims)
            result.update({v: val_reduced})

        return result

    @partial(jax.jit, static_argnames=["self"])
    def preprocess_parameters(self, parameters, num_backend: ModuleType = jnp):
        return super().preprocess_parameters(parameters, num_backend)
    
    @partial(jax.jit, static_argnames=["self"])
    def preprocess_x_in(self, x_in, num_backend: ModuleType = jnp):
        return super().preprocess_x_in(x_in, num_backend)

    @partial(jax.jit, static_argnames=["self"])
    def preprocess_y_0(self, y0, num_backend: ModuleType = jnp):
        return super().preprocess_y_0(y0, num_backend)

    @partial(jax.jit, static_argnames=["self"])
    def odesolve(self, y0, args, x_in):
        f = lambda t, y, args: self.model(t, y, *args)
        
        if len(x_in) > 0:
            if len(x_in) > 2:
                raise NotImplementedError(
                    "Currently only one interpolation is implemented, but "+
                    "it should be relatively simple to implement multiple "+
                    "interpolations. I assume, the interpolations could be "+
                    "passed as a list and expanded in the model. If you are "+
                    "dealing with this. Try pre-compute the interpolations. "+
                    "This should speed up the solver. "
                )
            
            
            if x_in[0].shape[0] != x_in[1].shape[0]:
                raise ValueError(
                    "Mismatch in zero-th dimensions of x and y in interpolation "+
                    "input 'x_in'. This often results of a problematic dimensional "+
                    "order. Consider reordering the dimensions and reordering the "+
                    "x dimension (e.g. time) after the batch dimension and before "+
                    "any other dimension."
                )
            interp = LinearInterpolation(ts=x_in[0], ys=x_in[1])
            args=(interp, *args)
            # jumps = x_in[0][self.coordinates_input_vars["x_in"][self.x_dim] < self.x[-1]]
            jumps = jnp.array(self.x_in_jumps, dtype=float)
        else:
            interp = None
            args=args
            jumps = None

        term = ODETerm(f)
        solver = self.diffrax_solver() # type: ignore (diffrax_solver is ensured
                                       # to be _MetaAbstractSolver type during 
                                       # post_init)
        saveat = SaveAt(ts=self.x)
        t_min = self.x[0]
        t_max = self.x[-1]
        # jump only those ts that are smaller than the last observations
        stepsize_controller = PIDController(
            rtol=self.rtol, atol=self.atol,
            pcoeff=self.pcoeff, icoeff=self.icoeff, dcoeff=self.dcoeff, 
            jump_ts=jumps,
        )
        
        sol = diffeqsolve(
            terms=term, 
            solver=solver, 
            t0=t_min, 
            t1=t_max, 
            dt0=0.1, 
            y0=tuple(y0), 
            args=args, 
            saveat=saveat, 
            stepsize_controller=stepsize_controller,
            adjoint=RecursiveCheckpointAdjoint(),
            max_steps=int(self.max_steps),
            # throw=False returns inf for all t > t_b, where t_b is the time 
            # at which the solver broke due to reaching max_steps. This behavior
            # happens instead of throwing an exception.
            throw=self.throw_exception
        )
        
        return tuple(sol.ys), interp

    @partial(jax.jit, static_argnames=["self", "odestates", "n_odeargs", "n_ppargs", "n_xin"])
    def odesolve_splitargs(self, *args, odestates, n_odeargs, n_ppargs, n_xin):
        n_odestates = len(odestates)
        y0 = args[:n_odestates]
        odeargs = args[n_odestates:n_odeargs+n_odestates]
        ppargs = args[n_odeargs+n_odestates:n_odeargs+n_odestates+n_ppargs]
        x_in = args[n_odestates+n_odeargs+n_ppargs:n_odestates+n_odeargs+n_ppargs+n_xin]
        sol, interp = self.odesolve(y0=y0, args=odeargs, x_in=x_in)
        
        res_dict = OrderedDict({v:val for v, val in zip(odestates, sol)})

        return self.post_processing(res_dict, jnp.array(self.x), interp, *ppargs)



import pytest
import numpy as np
import xarray as xr

from tests.fixtures import (
    create_simulation_for_test_numpyro_behavior,
    init_case_study_and_scenario
)

def replicate_obs(sim):
    sim.observations.expand_dims("id")
    obs1 = sim.observations.assign_coords({"id": 0})
    obs2 = sim.observations.assign_coords({"id": 1})
    obs = xr.combine_nested([obs1,obs2], concat_dim="id")
    sim.config.data_structure.rabbits.dimensions=["id", "time"]
    sim.config.data_structure.wolves.dimensions=["id", "time"]
    sim.config.simulation.batch_dimension = "id"
    sim.observations = obs

    sim.model_parameters["y0"] = sim.parse_input("y0", reference_data=obs, drop_dims=["time"])


# create config file for all following tests
create_simulation_for_test_numpyro_behavior()

def test_effect_handler_scaling():
    sim = init_case_study_and_scenario(
        case_study="lotka_volterra_case_study",
        scenario="test_numpyro_behavior"
    )
    sim.dispatch_constructor()
    sim.set_inferer("numpyro")
    idata = sim.inferer.prior_predictions(n=1)
    
    prior = idata.prior.sel(chain=0, draw=0)
    data_vars = sim.config.data_structure.observed_data_variables
    theta = {k: np.array(v.to_dict()["data"]) for k, v in prior.data_vars.items()}
    
    ll_sum_idata = idata.log_likelihood.sel(chain=0, draw=0)\
        .sum().to_array().sel(variable=data_vars)

    loglik_fn, _ = sim.inferer.create_log_likelihood(
        return_type="summed-by-site", scaled=False
    )

    loglik_fn_scaled, _ = sim.inferer.create_log_likelihood(
        return_type="summed-by-site", scaled=True
    )
    
    _, _, ll_sum_site = loglik_fn(theta=theta)
    _, _, ll_sum_site_scaled = loglik_fn_scaled(theta=theta)


    np.testing.assert_allclose(
        ll_sum_idata.values, 
        [ll_sum_site[f"{k}_obs"] for k in data_vars],
        atol=0.01,
        rtol=1e-7
    )

    # test if the scaled log-likelihood is equivalent to the scaled log-likelihood
    scales = {"rabbits": 2, "wolves": 1}
    np.testing.assert_allclose(
        ll_sum_idata.values, 
        [ll_sum_site_scaled[f"{k}_obs"] / scales[k] for k in data_vars],
        atol=0.01,
        rtol=1e-7
    )

def test_effect_handler_plate():
    import jax
    import jax.numpy as jnp
    import numpyro
    import numpyro.distributions as dist
    from numpyro.infer import MCMC, NUTS

    # Define the model with a multinomial inside a plate, using different n_trials for each batch
    def model(lethality, n_trials):
        batch_size, num_categories = lethality.shape  # Assumes probs has shape (batch_size, num_categories)
        probs = numpyro.sample("p", dist.Dirichlet(jnp.ones((batch_size,3))))
        
        # Define a plate for the batch dimension
        with numpyro.plate("batch", batch_size):
            # Sample from a Multinomial distribution with a different total_count for each batch item
            counts = numpyro.sample("lethality", dist.Multinomial(
                total_count=n_trials, probs=probs), obs=lethality)
        return counts

    # Define parameters for the multinomial distribution
    n_trials = np.array([10, 15, 20, 25, 30, 35, 40, 45, 50, 55]) * 100
    probs = np.array([
        [0.2, 0.5, 0.3],
        [0.3, 0.4, 0.3],
        [0.1, 0.7, 0.2],
        [0.25, 0.25, 0.5],
        [0.6, 0.2, 0.2],
        [0.3, 0.3, 0.4],
        [0.4, 0.4, 0.2],
        [0.2, 0.3, 0.5],
        [0.3, 0.3, 0.4],
        [0.1, 0.6, 0.3]
    ])  # Probability vector for each batch item

    lethality = jnp.array(list(map(lambda n, p: np.random.multinomial(n=n, pvals=p), n_trials, probs)))

    # Run MCMC sampling on the model
    nuts_kernel = NUTS(model)
    mcmc = MCMC(nuts_kernel, num_warmup=500, num_samples=1000)
    mcmc.run(jax.random.PRNGKey(0), lethality=lethality, n_trials=n_trials)

    # Get samples
    samples = mcmc.get_samples()["p"]

    # allow 5% deviation from any true probability value
    np.testing.assert_allclose(samples.mean(axis=0), probs, atol=0.05)


def test_effect_handler_reparam():
    pytest.skip()
    # TODO: Reparameterization can also be handled https://num.pyro.ai/en/stable/reparam.html#module-numpyro.infer.reparam

if __name__ == "__main__":
    pass
    # test_effect_handler_plate()
    # test_effect_handler_scaling()